"""Code for reading, writing, and representing less complicated praat data files

see **examples/get_vowel_points.py**
"""

import io
from typing import Tuple

from praatio.data_classes.data_point import PointObject1D, PointObject2D


def open1DPointObject(fn: str) -> PointObject1D:
    with io.open(fn, "r", encoding="utf-8") as fd:
        data = fd.read()
    if "xmin" in data[:100]:  # Kindof lazy
        data, objectType, minT, maxT = _parseNormalHeader(fn)

        start = 0
        dataList = []
        while True:
            try:
                start = data.index("=", start)
            except ValueError:
                break

            pointVal, start = _getNextValue(data, start)
            dataList.append((float(pointVal),))

        po = PointObject1D(dataList, objectType, minT, maxT)

    else:
        data, objectType, minT, maxT = _parseShortHeader(fn)
        dataList = [(float(val),) for val in data.split("\n") if val.strip() != ""]
        po = PointObject1D(dataList, objectType, minT, maxT)

    return po


def open2DPointObject(fn: str) -> PointObject2D:
    with io.open(fn, "r", encoding="utf-8") as fd:
        data = fd.read()
    if "xmin" in data[:100]:  # Kindof lazy
        data, objectType, minT, maxT = _parseNormalHeader(fn)

        start = 0
        dataList = []
        while True:
            try:
                start = data.index("=", start)
            except ValueError:
                break

            timeVal, start = _getNextValue(data, start)

            try:
                start = data.index("=", start)
            except ValueError:
                break

            pointVal, start = _getNextValue(data, start)
            dataList.append(
                (
                    float(timeVal),
                    float(pointVal),
                )
            )

        po = PointObject2D(dataList, objectType, minT, maxT)

    else:
        data, objectType, minT, maxT = _parseShortHeader(fn)
        dataStrList = data.split("\n")
        dataList = [
            (float(dataStrList[i]), float(dataStrList[i + 1]))
            for i in range(0, len(dataStrList), 2)
            if dataStrList[i].strip() != ""
        ]
        po = PointObject2D(dataList, objectType, minT, maxT)

    return po


def _parseNormalHeader(fn: str) -> Tuple[str, str, float, float]:
    with io.open(fn, "r", encoding="utf-8") as fd:
        data = fd.read()

    chunkedData = data.split("\n", 7)

    objectType = chunkedData[1].split("=")[-1]
    objectType = objectType.replace('"', "").strip()

    data = chunkedData[-1]
    maxT = float(chunkedData[-4].split("=")[-1].strip())
    minT = float(chunkedData[-5].split("=")[-1].strip())

    return data, objectType, minT, maxT


def _getNextValue(data: str, start: int) -> Tuple[str, int]:
    end = data.index("\n", start)
    value = data[start + 1 : end]
    return value, end


def _parseShortHeader(fn: str) -> Tuple[str, str, float, float]:
    with io.open(fn, "r", encoding="utf-8") as fd:
        data = fd.read()

    chunkedData = data.split("\n", 6)

    objectType = chunkedData[1].split("=")[-1]
    objectType = objectType.replace('"', "").strip()

    data = chunkedData[-1]
    maxT = float(chunkedData[-3])
    minT = float(chunkedData[-4])

    return data, objectType, minT, maxT



"""
A PointTier is a tier containing an array of points -- data that exists at a specific point in time
"""
from typing import List, Tuple, Optional, Any, Sequence

from typing_extensions import Literal

from praatio.utilities.constants import (
    Point,
    POINT_TIER,
)
from praatio.utilities import constants
from praatio.utilities import errors
from praatio.utilities import utils

from praatio.data_classes import textgrid_tier


def _homogenizeEntries(entries):
    """
    Enforces consistency in points

    - converts all entries to points
    - removes whitespace in labels
    - sorts values by time
    """
    processedEntries = [Point(float(time), label.strip()) for time, label in entries]
    processedEntries.sort()
    return processedEntries


def _calculateMinAndMaxTime(entries: Sequence[Point], minT=None, maxT=None):
    timeList = [time for time, label in entries]
    if minT is not None:
        timeList.append(float(minT))
    if maxT is not None:
        timeList.append(float(maxT))

    try:
        calculatedMinT = min(timeList)
        calculatedMaxT = max(timeList)
    except ValueError:
        raise errors.TimelessTextgridTierException()

    return (calculatedMinT, calculatedMaxT)


class PointTier(textgrid_tier.TextgridTier):

    tierType = POINT_TIER
    entryType = Point

    def __init__(
        self,
        name: str,
        entries: List[Point],
        minT: Optional[float] = None,
        maxT: Optional[float] = None,
    ):
        """A point tier is for annotating instaneous events

        The entries is of the form:
        [(timeVal1, label1), (timeVal2, label2), ]

        The data stored in the labels can be anything but will
        be interpreted as text by praatio (the label could be descriptive
        text e.g. ('peak point here') or numerical data e.g. (pitch values
        like '132'))
        """
        entries = _homogenizeEntries(entries)
        calculatedMinT, calculatedMaxT = _calculateMinAndMaxTime(entries, minT, maxT)

        super(PointTier, self).__init__(name, entries, calculatedMinT, calculatedMaxT)

    def crop(
        self,
        cropStart: float,
        cropEnd: float,
        mode: Literal["strict", "lax", "truncated"] = "lax",
        rebaseToZero: bool = True,
    ) -> "PointTier":
        """Creates a new tier containing all entries inside the new interval

        Args:
            cropStart:
            cropEnd:
            mode: Mode is ignored.  This parameter is kept for
                compatibility with IntervalTier.crop()
            rebaseToZero: if True, all entries will have their
                timestamps subtracted by *cropStart*.

        Returns:
            the modified version of the current tier
        """
        if cropStart >= cropEnd:
            raise errors.ArgumentError(
                f"Crop error: start time ({cropStart}) must occur before end time ({cropEnd})"
            )

        newEntries = []

        for entry in self.entries:
            timestamp = entry.time

            if timestamp >= cropStart and timestamp <= cropEnd:
                newEntries.append(entry)

        if rebaseToZero is True:
            newEntries = [
                Point(timeV - cropStart, label) for timeV, label in newEntries
            ]
            minT = 0.0
            maxT = cropEnd - cropStart
        else:
            minT = cropStart
            maxT = cropEnd

        return PointTier(self.name, newEntries, minT, maxT)

    def deleteEntry(self, entry: Point) -> None:
        """Removes an entry from the entries"""
        self._entries.pop(self._entries.index(entry))

    def editTimestamps(
        self,
        offset: float,
        reportingMode: Literal["silence", "warning", "error"] = "warning",
    ) -> "PointTier":
        """Modifies all timestamps by a constant amount

        Args:
            offset:
            reportingMode: one of "silence", "warning", or "error". This flag
                determines the behavior if an entries moves outside of minTimestamp
                or maxTimestamp after being edited

        Returns:
            the modified version of the current tier
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )
        errorReporter = utils.getErrorReporter(reportingMode)

        newEntries: List[Point] = []
        for timestamp, label in self.entries:

            newTimestamp = timestamp + offset
            utils.checkIsUndershoot(newTimestamp, self.minTimestamp, errorReporter)
            utils.checkIsOvershoot(newTimestamp, self.maxTimestamp, errorReporter)

            if newTimestamp < 0:
                continue

            newEntries.append(Point(newTimestamp, label))

        # Determine new min and max timestamps
        timeList = [float(point.time) for point in newEntries]
        newMin = min(timeList)
        newMax = max(timeList)

        if newMin > self.minTimestamp:
            newMin = self.minTimestamp

        if newMax < self.maxTimestamp:
            newMax = self.maxTimestamp

        return PointTier(self.name, newEntries, newMin, newMax)

    def getValuesAtPoints(
        self,
        dataTupleList: List[Tuple[float, ...]],
        fuzzyMatching: bool = False,
    ) -> List[Tuple[Any, ...]]:
        """Get the values that occur at points in the point tier

        The procedure assumes that all data is ordered in time.
        dataTupleList should be in the form
        [(t1, v1a, v1b, ..), (t2, v2a, v2b, ..), ..]

        It returns the data in the form of
        [(t1, v1a, v1b, ..), (t2, v2a, v2b), ..]

        The procedure makes one pass through dataTupleList and one
        pass through self.entries.  If the data is not sequentially
        ordered, the incorrect response will be returned.

        Args:
            dataTupleList:
            fuzzyMatching: if True, if there is not a feature value
                at a point, the nearest feature value will be taken.

        Returns:
            A list of values that exist at the given timepoints
        """

        currentIndex = 0
        retList = []

        sortedDataTupleList = sorted(dataTupleList)
        for timestamp, label in self.entries:
            retTuple = utils.getValueAtTime(
                timestamp,
                sortedDataTupleList,
                fuzzyMatching=fuzzyMatching,
                startI=currentIndex,
            )
            retRow, currentIndex = retTuple
            retList.append(retRow)

        return retList

    def eraseRegion(
        self,
        start: float,
        end: float,
        collisionMode: Literal["truncate", "categorical", "error"] = "error",
        doShrink: bool = True,
    ) -> "PointTier":
        """Makes a region in a tier blank (removes all contained entries)

        Args:
            start: the start of the deletion interval
            end: the end of the deletion interval
            collisionMode: Ignored for the moment (added for compatibility with
                eraseRegion() for Interval Tiers)
            doShrink: if True, moves leftward by (/end/ - /start/) all points
                to the right of /end/

        Returns:
            The modified version of the current tier
        """

        newTier = self.new()
        croppedTier = newTier.crop(start, end, constants.CropCollision.TRUNCATED, False)
        matchList = croppedTier.entries

        if len(matchList) > 0:
            # Remove all the matches from the entries
            # Go in reverse order because we're destructively altering
            # the order of the list (messes up index order)
            for point in matchList[::-1]:
                newTier.deleteEntry(point)

        if doShrink is True:
            newEntries = []
            diff = end - start
            for point in newTier.entries:
                if point.time < start:
                    newEntries.append(point)
                elif point.time > end:
                    newEntries.append(Point(point.time - diff, point.label))

            newMax = newTier.maxTimestamp - diff
            newTier = newTier.new(entries=newEntries, maxTimestamp=newMax)

        return newTier

    def insertEntry(
        self,
        entry: Point,
        collisionMode: Literal["replace", "merge", "error"] = "error",
        collisionReportingMode: Literal["silence", "warning"] = "warning",
    ) -> None:
        """Inserts an interval into the tier

        Args:
            entry: the entry to insert
            collisionMode: determines the behavior if intervals exist in
                the insertion area.
                - 'replace', existing items will be removed
                - 'merge', inserting item will be fused with existing items
                - 'error', will throw TextgridCollisionException
            collisionReportingMode:

        Returns:
            None
        """

        utils.validateOption(
            "collisionMode", collisionMode, constants.IntervalCollision
        )
        utils.validateOption(
            "collisionReportingMode",
            collisionReportingMode,
            constants.ErrorReportingMode,
        )
        collisionReporter = utils.getErrorReporter(collisionReportingMode)

        if not isinstance(entry, Point):
            newPoint = Point(entry[0], entry[1])
        else:
            newPoint = entry

        matchList = []
        i = None
        for i, point in enumerate(self.entries):
            if point.time == newPoint.time:
                matchList.append(point)
                break

        if len(matchList) == 0:
            self._entries.append(newPoint)

        elif collisionMode == constants.IntervalCollision.REPLACE:
            self.deleteEntry(self.entries[i])
            self._entries.append(newPoint)

        elif collisionMode == constants.IntervalCollision.MERGE:
            oldPoint = self.entries[i]
            mergedPoint = Point(
                newPoint.time, "-".join([oldPoint.label, newPoint.label])
            )
            self.deleteEntry(self._entries[i])
            self._entries.append(mergedPoint)

        else:
            raise errors.CollisionError(
                f"Attempted to insert interval {point} into tier {self.name} "
                "of textgrid but overlapping entries "
                f"{[tuple(interval) for interval in matchList]} "
                "already exist"
            )

        self.sort()

        if len(matchList) != 0:
            collisionReporter(
                errors.CollisionError,
                f"Collision warning for ({point}) with items ({matchList}) of tier '{self.name}'",
            )

    def insertSpace(
        self,
        start: float,
        duration: float,
        _collisionMode: Literal["stretch", "split", "no_change", "error"] = "error",
    ) -> "PointTier":
        """Inserts a region into the tier

        Args:
            start: the start time to insert a space at
            duration: the duration of the space to insert
            collisionMode: Ignored for the moment (added for compatibility
                with insertSpace() for Interval Tiers)

        Returns:
            PointTier: the modified version of the current tier
        """

        newEntries = []
        for point in self.entries:
            if point.time <= start:
                newEntries.append(point)
            elif point.time > start:
                newEntries.append(Point(point.time + duration, point.label))

        newTier = self.new(
            entries=newEntries, maxTimestamp=self.maxTimestamp + duration
        )

        return newTier

    def validate(
        self, reportingMode: Literal["silence", "warning", "error"] = "warning"
    ) -> bool:
        """Validate this tier

        Returns whether the tier is valid or not. If reportingMode is "warning"
        or "error" this will also print on error or stop execution, respectively.

        Args:
            reportingMode: Determines the behavior if there is a size difference
                between the maxTimestamp in the tier and the current textgrid.

        Returns:
            True if this tier is valid; False otherwise
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )
        errorReporter = utils.getErrorReporter(reportingMode)

        isValid = True
        previousPoint = None
        for point in self.entries:
            if previousPoint and previousPoint.time > point.time:
                isValid = False
                errorReporter(
                    errors.TextgridStateError,
                    f"Points are not sorted in time: "
                    f"[({previousPoint}), ({point})]",
                )

            if utils.checkIsUndershoot(point.time, self.minTimestamp, errorReporter):
                isValid = False

            if utils.checkIsOvershoot(point.time, self.maxTimestamp, errorReporter):
                isValid = False

            previousPoint = point

        return isValid



"""
Functions for reading/writing/manipulating textgrid files.

This file links the main data structures for representing Textgrid data:
Textgrid, IntervalTier, and PointTier

A Textgrid is a container for multiple annotation tiers.  Tiers can contain
either interval data (IntervalTier) or point data (PointTier).
Tiers in a Textgrid are ordered and must contain a unique name.

openTextgrid() can be used to open a textgrid file.
Textgrid.save() can be used to save a Textgrid object to a file.

Historically, these three classes lived in this file. To
make maintance easier, they were split off.  You can still access
them from this file as before, but please check out the respective
files for more documentation on each class:
IntervalTier in data_classes/interval_tier.py
PointTier in data_classes/point_tier.py
Textgrid in data_classes/textgrid.py

see the **examples/** directory for examples using textgrid.py
"""

import io
from typing import (
    Union,
    Type,
)

from typing_extensions import Literal


from praatio.utilities.constants import (
    INTERVAL_TIER,
)
from praatio.data_classes.interval_tier import IntervalTier
from praatio.data_classes.point_tier import PointTier
from praatio.data_classes.textgrid import Textgrid
from praatio.utilities import textgrid_io
from praatio.utilities import utils
from praatio.utilities import constants
from praatio.utilities import errors


def openTextgrid(
    fnFullPath: str,
    includeEmptyIntervals: bool,
    reportingMode: Literal["silence", "warning", "error"] = "warning",
    duplicateNamesMode: Literal["error", "rename"] = "error",
) -> Textgrid:
    """Opens a textgrid file (.TextGrid and .json are both fine)

    https://www.fon.hum.uva.nl/praat/manual/TextGrid_file_formats.html

    Args:
        fnFullPath: the path to the textgrid to open
        includeEmptyIntervals: if False, points and intervals with
             an empty label '' are not included in the returned Textgrid

    Returns:
        A Textgrid

    Raises:
        DuplicateTierName: The textgrid contains multiple tiers with the same name
    """
    utils.validateOption("reportingMode", reportingMode, constants.ErrorReportingMode)
    utils.validateOption(
        "duplicateNamesMode", duplicateNamesMode, constants.DuplicateNames
    )
    try:
        with io.open(fnFullPath, "r", encoding="utf-16") as fd:
            data = fd.read()
    except UnicodeError:
        with io.open(fnFullPath, "r", encoding="utf-8") as fd:
            data = fd.read()

    tgAsDict = textgrid_io.parseTextgridStr(data, includeEmptyIntervals)

    tierNames = []
    for tier in tgAsDict["tiers"]:
        name = tier["name"]
        if name in tierNames:
            if duplicateNamesMode == constants.DuplicateNames.ERROR:
                raise errors.DuplicateTierName(
                    f"Your textgrid contains tiers with the same name '{name}'. "
                    "This is not allowed. It is recommended that you rename them. "
                    "If you set openTextgrid(..., duplicateNamesMode='rename'), praatio "
                    "will automatically append numbers to the end of tiers to ensure they "
                    "are unique."
                )
            elif duplicateNamesMode == constants.DuplicateNames.RENAME:
                newName = name
                i = 2
                while newName in tierNames:
                    newName = f"{name}_{i}"
                    i += 1
                name = newName
                tier["name"] = name
        tierNames.append(name)

    return _dictionaryToTg(tgAsDict, reportingMode)


def _dictionaryToTg(
    tgAsDict: dict, reportingMode: Literal["silence", "warning", "error"]
) -> Textgrid:
    """Converts a dictionary representation of a textgrid to a Textgrid"""
    utils.validateOption("reportingMode", reportingMode, constants.ErrorReportingMode)

    tg = Textgrid()
    tg.minTimestamp = tgAsDict["xmin"]
    tg.maxTimestamp = tgAsDict["xmax"]

    for tierAsDict in tgAsDict["tiers"]:
        klass: Union[Type[PointTier], Type[IntervalTier]]
        if tierAsDict["class"] == INTERVAL_TIER:
            klass = IntervalTier
        else:
            klass = PointTier
        tier = klass(
            tierAsDict["name"],
            tierAsDict["entries"],
            tierAsDict["xmin"],
            tierAsDict["xmax"],
        )
        tg.addTier(tier, reportingMode=reportingMode)

    return tg



# coding: utf-8
"""Functions for working with pitch data

This file depends on the praat script get_pitch_and_intensity.praat
(which depends on praat) to extract pitch and intensity values from
audio data.  Once the data is extracted, there are functions for
data normalization and calculating various measures from the time
stamped output of the praat script (ie **generatePIMeasures()**)

For brevity, 'pitch_and_intensity' is referred to as 'PI'

see **examples/get_pitch_and_formants.py**
"""

import os
from os.path import join
import io
import math
from typing import List, Tuple, Optional, cast

from praatio import data_points
from praatio import praatio_scripts
from praatio import textgrid
from praatio.utilities import errors
from praatio.utilities import my_math
from praatio.utilities import utils
from praatio.utilities.constants import Point


HERTZ = "Hertz"
UNSPECIFIED = "unspecified"
_PITCH_ERROR_TIER_NAME = "pitch errors"


def _extractPIPiecewise(
    inputFN: str,
    outputFN: str,
    praatEXE: str,
    minPitch: float,
    maxPitch: float,
    tgFN: str,
    tierName: str,
    tmpOutputPath: str,
    sampleStep: float = 0.01,
    silenceThreshold: float = 0.03,
    pitchUnit: str = HERTZ,
    forceRegenerate: bool = True,
    undefinedValue: float = None,
    medianFilterWindowSize: int = 0,
    pitchQuadInterp: bool = False,
) -> List[Tuple[float, ...]]:
    """Extracts pitch and int from each labeled interval in a textgrid

    This has the benefit of being faster than using _extractPIFile if only
    labeled regions need to have their pitch values sampled, particularly
    for longer files.

    Returns the result as a list.  Will load the serialized result
    if this has already been called on the appropriate files before
    """
    outputPath = os.path.split(outputFN)[0]
    utils.makeDir(outputPath)

    windowSize = medianFilterWindowSize

    if not os.path.exists(inputFN):
        raise errors.ArgumentError(f"Required folder does not exist: f{inputFN}")

    firstTime = not os.path.exists(outputFN)
    if firstTime or forceRegenerate is True:

        utils.makeDir(tmpOutputPath)
        splitAudioList = praatio_scripts.splitAudioOnTier(
            inputFN, tgFN, tierName, tmpOutputPath, False
        )
        allPIList: List[Tuple[str, str, str]] = []
        for start, _, fn in splitAudioList:
            tmpTrackName = os.path.splitext(fn)[0] + ".txt"
            piList = _extractPIFile(
                join(tmpOutputPath, fn),
                join(tmpOutputPath, tmpTrackName),
                praatEXE,
                minPitch,
                maxPitch,
                sampleStep,
                silenceThreshold,
                pitchUnit,
                forceRegenerate=True,
                medianFilterWindowSize=windowSize,
                pitchQuadInterp=pitchQuadInterp,
            )
            convertedPiList = [
                ("%0.3f" % (float(time) + start), str(pV), str(iV))
                for time, pV, iV in piList
            ]
            allPIList.extend(convertedPiList)

        outputData = [",".join(row) for row in allPIList]
        with open(outputFN, "w") as fd:
            fd.write("\n".join(outputData) + "\n")

    return loadTimeSeriesData(outputFN, undefinedValue=undefinedValue)


def _extractPIFile(
    inputFN: str,
    outputFN: str,
    praatEXE: str,
    minPitch: float,
    maxPitch: float,
    sampleStep: float = 0.01,
    silenceThreshold: float = 0.03,
    pitchUnit: str = HERTZ,
    forceRegenerate: bool = True,
    undefinedValue: float = None,
    medianFilterWindowSize: int = 0,
    pitchQuadInterp: bool = False,
) -> List[Tuple[float, ...]]:
    """Extracts pitch and intensity values from an audio file

    Returns the result as a list.  Will load the serialized result
    if this has already been called on the appropriate files before
    """
    outputPath = os.path.split(outputFN)[0]
    utils.makeDir(outputPath)

    if not os.path.exists(inputFN):
        raise errors.ArgumentError(f"Required folder does not exist: f{inputFN}")

    firstTime = not os.path.exists(outputFN)
    if firstTime or forceRegenerate is True:

        # The praat script uses append mode, so we need to clear any prior
        # result
        if os.path.exists(outputFN):
            os.remove(outputFN)

        if pitchQuadInterp is True:
            doInterpolation = 1
        else:
            doInterpolation = 0

        argList = [
            inputFN,
            outputFN,
            sampleStep,
            minPitch,
            maxPitch,
            silenceThreshold,
            pitchUnit,
            -1,
            -1,
            medianFilterWindowSize,
            doInterpolation,
        ]

        scriptName = "get_pitch_and_intensity.praat"
        scriptFN = join(utils.scriptsPath, scriptName)
        utils.runPraatScript(praatEXE, scriptFN, argList)

    return loadTimeSeriesData(outputFN, undefinedValue=undefinedValue)


def extractIntensity(
    inputFN: str,
    outputFN: str,
    praatEXE: str,
    minPitch: float,
    sampleStep: float = 0.01,
    forceRegenerate: bool = True,
    undefinedValue: float = None,
) -> List[Tuple[float, ...]]:
    """Extract the intensity for an audio file

    Calculates intensity using the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_Intensity___.html
    """
    outputPath = os.path.split(outputFN)[0]
    utils.makeDir(outputPath)

    if not os.path.exists(inputFN):
        raise errors.ArgumentError(f"Required folder does not exist: f{inputFN}")

    firstTime = not os.path.exists(outputFN)
    if firstTime or forceRegenerate is True:

        # The praat script uses append mode, so we need to clear any prior
        # result
        if os.path.exists(outputFN):
            os.remove(outputFN)

        argList = [inputFN, outputFN, sampleStep, minPitch, -1, -1]

        scriptName = "get_intensity.praat"
        scriptFN = join(utils.scriptsPath, scriptName)
        utils.runPraatScript(praatEXE, scriptFN, argList)

    return loadTimeSeriesData(outputFN, undefinedValue=undefinedValue)


def extractPitchTier(
    wavFN: str,
    outputFN: str,
    praatEXE: str,
    minPitch: float,
    maxPitch: float,
    sampleStep: float = 0.01,
    silenceThreshold: float = 0.03,
    forceRegenerate: bool = True,
    medianFilterWindowSize: int = 0,
    pitchQuadInterp: bool = False,
) -> data_points.PointObject2D:
    """Extract pitch at regular intervals from the input wav file

    Data is output to a text file and then returned in a list in the form
    [(timeV1, pitchV1), (timeV2, pitchV2), ...]

    Calculates pitch using the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_Pitch___.html

    Args:
        sampleStep: the frequency to sample pitch at
        silenceThreshold: segments with lower intensity won't be analyzed
            for pitch
        forceRegenerate: if running this function for the same file, if False
            just read in the existing pitch file
        pitchQuadInterp: if True, quadratically interpolate pitch

    Returns:
        The pitch data
    """
    outputPath = os.path.split(outputFN)[0]

    utils.makeDir(outputPath)

    if pitchQuadInterp is True:
        doInterpolation = 1
    else:
        doInterpolation = 0

    if not os.path.exists(wavFN):
        raise errors.ArgumentError(f"Required file does not exist: f{wavFN}")

    firstTime = not os.path.exists(outputFN)
    if firstTime or forceRegenerate is True:
        if os.path.exists(outputFN):
            os.remove(outputFN)

        argList = [
            wavFN,
            outputFN,
            sampleStep,
            minPitch,
            maxPitch,
            silenceThreshold,
            medianFilterWindowSize,
            doInterpolation,
        ]

        scriptName = "get_pitchtier.praat"
        scriptFN = join(utils.scriptsPath, scriptName)
        utils.runPraatScript(praatEXE, scriptFN, argList)

    return data_points.open2DPointObject(outputFN)


def extractPitch(
    wavFN: str,
    outputFN: str,
    praatEXE: str,
    minPitch: float,
    maxPitch: float,
    sampleStep: float = 0.01,
    silenceThreshold: float = 0.03,
    forceRegenerate: bool = True,
    undefinedValue: float = None,
    medianFilterWindowSize: int = 0,
    pitchQuadInterp: bool = False,
) -> List[Tuple[float, ...]]:
    """Extract pitch at regular intervals from the input wav file

    Data is output to a text file and then returned in a list in the form
    [(timeV1, pitchV1), (timeV2, pitchV2), ...]

    Calculates pitch using the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_Pitch___.html

    Args:
        sampleStep - the frequency to sample pitch at
        silenceThreshold - segments with lower intensity won't be analyzed
            for pitch
        forceRegenerate - if running this function for the same file, if False
            just read in the existing pitch file
        undefinedValue - if None remove from the dataset, otherset set to
            undefinedValue
        pitchQuadInterp - if True, quadratically interpolate pitch
    """
    outputPath = os.path.split(outputFN)[0]

    utils.makeDir(outputPath)

    if pitchQuadInterp is True:
        doInterpolation = 1
    else:
        doInterpolation = 0

    if not os.path.exists(wavFN):
        raise errors.ArgumentError(f"Required file does not exist: f{wavFN}")

    firstTime = not os.path.exists(outputFN)
    if firstTime or forceRegenerate is True:
        if os.path.exists(outputFN):
            os.remove(outputFN)

        argList = [
            wavFN,
            outputFN,
            sampleStep,
            minPitch,
            maxPitch,
            silenceThreshold,
            -1,
            -1,
            medianFilterWindowSize,
            doInterpolation,
        ]

        scriptName = "get_pitch.praat"
        scriptFN = join(utils.scriptsPath, scriptName)
        utils.runPraatScript(praatEXE, scriptFN, argList)

    return loadTimeSeriesData(outputFN, undefinedValue=undefinedValue)


def extractPI(
    inputFN: str,
    outputFN: str,
    praatEXE: str,
    minPitch: float,
    maxPitch: float,
    sampleStep: float = 0.01,
    silenceThreshold: float = 0.03,
    pitchUnit: str = HERTZ,
    forceRegenerate: bool = True,
    tgFN: str = None,
    tierName: str = None,
    tmpOutputPath: str = None,
    undefinedValue: float = None,
    medianFilterWindowSize: int = 0,
    pitchQuadInterp: bool = False,
) -> List[Tuple[float, ...]]:
    """Extracts pitch and intensity from a file wholesale or piecewise

    If the parameters for a tg are passed in, this will only extract labeled
    segments in a tier of the tg.  Otherwise, pitch will be extracted from
    the entire file.

    male: minPitch=50; maxPitch=350
    female: minPitch=75; maxPitch=450
    pitchUnit: "Hertz", "semitones re 100 Hz", etc

    Calculates pitch and intensity using the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_Pitch___.html
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_Intensity___.html
    """

    outputPath = os.path.split(outputFN)[0]

    windowSize = medianFilterWindowSize

    if tgFN is None or tierName is None:
        piList = _extractPIFile(
            inputFN,
            outputFN,
            praatEXE,
            minPitch,
            maxPitch,
            sampleStep,
            silenceThreshold,
            pitchUnit,
            forceRegenerate,
            undefinedValue=undefinedValue,
            medianFilterWindowSize=windowSize,
            pitchQuadInterp=pitchQuadInterp,
        )
    else:
        if tmpOutputPath is None:
            tmpOutputPath = join(outputPath, "piecewise_output")
        piList = _extractPIPiecewise(
            inputFN,
            outputFN,
            praatEXE,
            minPitch,
            maxPitch,
            tgFN,
            tierName,
            tmpOutputPath,
            sampleStep,
            silenceThreshold,
            pitchUnit,
            forceRegenerate,
            undefinedValue=undefinedValue,
            medianFilterWindowSize=windowSize,
            pitchQuadInterp=pitchQuadInterp,
        )

    return piList


def loadTimeSeriesData(
    fn: str, undefinedValue: float = None
) -> List[Tuple[float, ...]]:
    """For reading the output of get_pitch_and_intensity or get_intensity

    Data should be of the form
    [(time1, value1a, value1b, ...),
     (time2, value2a, value2b, ...), ]
    """
    name = os.path.splitext(os.path.split(fn)[1])[0]

    try:
        with io.open(fn, "r", encoding="utf-8") as fd:
            data = fd.read()
    except IOError:
        print(f"No pitch track for: {name}")
        raise

    dataList = [row.split(",") for row in data.splitlines() if row != ""]

    # The new praat script includes a header
    if dataList[0][0] == "time":
        dataList = dataList[1:]

    newDataList = []
    for row in dataList:
        time = float(row.pop(0))
        entry = [
            time,
        ]
        doSkip = False
        for value in row:
            if "--" in value:
                if undefinedValue is not None:
                    appendValue = undefinedValue
                else:
                    doSkip = True
                    break
            else:
                appendValue = float(value)

            entry.append(appendValue)

        if doSkip is True:
            continue

        newDataList.append(tuple(entry))

    return newDataList


def generatePIMeasures(
    dataList: List[Tuple[float, float, float]],
    tgFN: str,
    tierName: str,
    doPitch: bool,
    medianFilterWindowSize: int = None,
    globalZNormalization: bool = False,
    localZNormalizationWindowSize: int = 0,
) -> List[Tuple[float, ...]]:
    """Generates processed values for the labeled intervals in a textgrid

    Args:
        doPitch: if True get pitch measures; if False get rms intensity
        medianFilterWindowSize: if none, no filtering is done
        globalZNormalization: if True, values are normalized with the mean
            and stdDev of the data in dataList
        localZNormalization: if greater than 1, values are normalized with the mean
            and stdDev of the local context (for a window of 5, it
            would consider the current value, 2 values before and 2
            values after)
    """

    # Warn user that normalizing a second time nullifies the first normalization
    if globalZNormalization is True and localZNormalizationWindowSize > 0:
        raise errors.NormalizationException()

    castDataList = cast(List[Tuple[float, ...]], dataList)
    if globalZNormalization is True:
        if doPitch:
            castDataList = my_math.znormalizeSpeakerData(castDataList, 1, True)
        else:
            castDataList = my_math.znormalizeSpeakerData(castDataList, 2, True)

    # Raw values should have 0 filtered; normalized values are centered around 0, so don't filter
    filterZeroFlag = not globalZNormalization

    tg = textgrid.openTextgrid(tgFN, False)
    if not isinstance(tg.getTier(tierName), textgrid.IntervalTier):
        raise errors.IncompatibleTierError(tg.getTier(tierName))

    tier = cast(textgrid.IntervalTier, tg.getTier(tierName))
    piData = tier.getValuesInIntervals(castDataList)

    outputList: List[List[float]] = []
    for interval, entries in piData:
        label = interval[0]
        if doPitch:
            tmpValList = [f0Val for _, f0Val, _ in entries]
            f0Measures = getPitchMeasures(
                tmpValList, tgFN, label, medianFilterWindowSize, filterZeroFlag
            )
            outputList.append(list(f0Measures))
        else:
            tmpValList = [intensityVal for _, _, intensityVal in entries]

            if filterZeroFlag:
                tmpValList = [
                    intensityVal for intensityVal in tmpValList if intensityVal != 0.0
                ]

            rmsIntensity = 0.0
            if len(tmpValList) != 0:
                rmsIntensity = my_math.rms(tmpValList)
            outputList.append(
                [
                    rmsIntensity,
                ]
            )

    # Locally normalize the output
    if localZNormalizationWindowSize > 0 and len(outputList) > 0:
        for colI in range(len(outputList[0])):
            featValList = [row[colI] for row in outputList]

            featValList = my_math.znormWindowFilter(
                featValList, localZNormalizationWindowSize, True, True
            )
            if len(featValList) != len(outputList):  # This should hopefully not happen
                raise errors.UnexpectedError(
                    "Lists must be of the same length but are not: "
                    f"({len(featValList)}), ({len(outputList)})"
                )

            for i, val in enumerate(featValList):
                outputList[i][colI] = val

    return [tuple(row) for row in outputList]


def getPitchMeasures(
    f0Values: List[float],
    name: str = None,
    label: str = None,
    medianFilterWindowSize: int = None,
    filterZeroFlag: bool = False,
) -> Tuple[float, float, float, float, float, float]:
    """Get various measures (min, max, etc) for the passed in list of pitch values

    Args:
        name: the name of the file.
        label: the label of the current interval.
        medianFilterWindowSize: if None, there is no median filtering
        filterZeroFlag: if True, values of zero are removed
    """

    if name is None:
        name = UNSPECIFIED
    if label is None:
        label = UNSPECIFIED

    if medianFilterWindowSize is not None:
        f0Values = my_math.medianFilter(
            f0Values, medianFilterWindowSize, useEdgePadding=True
        )

    if filterZeroFlag:
        f0Values = [f0Val for f0Val in f0Values if int(f0Val) != 0]

    if len(f0Values) == 0:
        myStr = f"No pitch data for file: {name}, label: {label}"
        print(myStr.encode("ascii", "replace"))
        counts = 0.0
        meanF0 = 0.0
        maxF0 = 0.0
        minF0 = 0.0
        rangeF0 = 0.0
        variance = 0.0
        std = 0.0
    else:
        counts = float(len(f0Values))
        meanF0 = sum(f0Values) / counts
        maxF0 = max(f0Values)
        minF0 = min(f0Values)
        rangeF0 = maxF0 - minF0

        variance = sum([(val - meanF0) ** 2 for val in f0Values]) / counts
        std = math.sqrt(variance)

    return (meanF0, maxF0, minF0, rangeF0, variance, std)


def detectPitchErrors(
    pitchList: List[Tuple[float, float]],
    maxJumpThreshold: float = 0.70,
    tgToMark: Optional[textgrid.Textgrid] = None,
) -> Tuple[List[Point], Optional[textgrid.Textgrid]]:
    """Detect pitch halving and doubling errors.

    If a textgrid is passed in, it adds the markings to the textgrid
    """
    if maxJumpThreshold < 0 or maxJumpThreshold > 1:
        raise errors.ArgumentError(
            f"'maxJumpThreshold' must be between 0 and 1.  Was given ({maxJumpThreshold})"
        )

    tierName = _PITCH_ERROR_TIER_NAME
    if tgToMark is not None and tierName in tgToMark.tierNames:
        raise errors.ArgumentError(
            f"Tier name '{tierName}' is already in provided textgrid"
        )

    errorList = []
    for i in range(1, len(pitchList)):
        lastPitch = pitchList[i - 1][1]
        currentPitch = pitchList[i][1]

        ceilingCutoff = currentPitch / maxJumpThreshold
        floorCutoff = currentPitch * maxJumpThreshold
        if (lastPitch <= floorCutoff) or (lastPitch >= ceilingCutoff):
            currentTime = pitchList[i][0]
            errorList.append(Point(currentTime, str(currentPitch / lastPitch)))

    if tgToMark is not None:
        pointTier = textgrid.PointTier(
            tierName, errorList, tgToMark.minTimestamp, tgToMark.maxTimestamp
        )
        tgToMark.addTier(pointTier)

    return errorList, tgToMark



class WrongPraatioVersion(Exception):
    pass


raise WrongPraatioVersion(
    "\n\nWARNING: You've tried to import 'tgio' which was renamed 'textgrid' in praatio 5.x.\n"
    "Many other, breaking changes were made.\n"
    "You will need to modify your code to use praatio 5.0.\n"
    "If you would like to use your code without changes, please run the following "
    "two instructions from the command line to install praatio 4.x:\n\n"
    "pip uninstall praatio\n"
    "pip install 'praatio<5'"
)



# encoding: utf-8
"""
This is a table of TIMIT-and-ARPABET-to-IPA codes.
"""

timitToIPA = {
    u"ng": u"ŋ",
    u"sh": u"ʃ",
    u"ch": u"tʃ",
    u"y": u"j",
    u"zj": u"ʒ",
    u"zh": u"ʒ",
    u"jh": u"dʒ",
    u"dh": u"ð",
    u"dx": u"ɾ",
    u"tq": u"ʔ",
    u"th": u"θ",
    u"nx": u"n",
    u"p": u"p",
    u"b": u"b",
    u"m": u"m",
    u"t": u"t",
    u"d": u"d",
    u"n": u"n",
    u"k": u"k",
    u"g": u"g",
    u"s": u"s",
    u"z": u"z",
    u"f": u"f",
    u"v": u"v",
    u"w": u"w",
    u"l": u"l",
    u"r": u"r",
    u"aa": u"ɑ",
    u"h": u"h",
    u"hh": u"h",
    u"ae": u"æ",
    u"ah": u"ʌ",
    u"ah0": u"ə",
    u"ah1": u"ʌ",
    u"ao": u"ɔ",
    u"eh": u"ɛ",
    u"ih": u"ɪ",
    u"iy": u"i",
    u"uh": u"ʊ",
    u"uw": u"u",
    u"ay": u"ɑɪ",
    u"aw": u"ɑʊ",
    u"ey": u"eɪ",
    u"ow": u"oʊ",
    u"oy": u"ɔɪ",
    u"er": u"ɚ",
    u"er0": u"ɚ",
    u"er1": u"ɝ",
    u"en": u"n̩",
    u"el": u"l̩",
    # Buckeye contains some vowel + consonant sequences
    u"ihn": u"ɪn",
    u"ayn": u"ɑɪn",
    u"uhn": u"ən",
    u"ehn": u"ɛn",
    u"ehm": u"ɛm",
    u"em": u"ɛm",
    u"aan": u"ɑn",
}



"""
Praatio is a library for working praat files and audio data.

Praat is a popular tool for working with transcribed speech data.
It has tools for annotating speech, manipulating speech, and extracting
information from speech.
Praat's homepage: [http://www.fon.hum.uva.nl/praat/](http://www.fon.hum.uva.nl/praat/)

praatio's utility comes from the classes Textgrid, IntervalTier,
and PointTier in **textgrid.py** These classes represent the speech transcript
data stored inside of .TextGrid files.  textgrid.py contains functions for
reading/writing these objects from/to files. **data_points.py** and **klattgrid.py**
contain similar code for other data files that praat works with.

**pitch_and_intensity.py** uses praat to extract pitch and intensity values
from audio using praat.

**praat_scripts.py** enables one to run various praat scripts from within python.
If preferred, one can access the **praatScripts/** directory to view and edit the
praat scripts or open them directly in praat. They can all be run independentlyof python.

**audio.py** is for reading, reading, and performing simple manipulations to wave files.

**praatio_scripts.py** contains code that combines various code in praatio--mostly
high-level functions for working with textgrids, sometimes in conjuction with audio files.

For example scripts that use various parts of the praatio library, please see 'examples/'
in the root directory. If you downloaded praatio from pip, you can get the example files
on the praatio github page:
[https://github.com/timmahrt/praatIO](https://github.com/timmahrt/praatIO)

There is also a tutorial for working with praatio on the github page.
"""

__all__ = [
    "audio",
    "data_points",
    "klattgrid",
    "pitch_and_intensity",
    "praat_scripts",
    "praatio_scripts",
    "textgrid",
    "data_classes",
    "utilities",
]



"""
Functions for reading, writing, querying, and manipulating audio.

see **examples/anonymize_recording.py**, **examples/delete_vowels.py**,
and **examples/extract_subwavs.py**
"""

import math
import wave
import struct
import copy
from typing import List, Tuple, Optional, Callable
from abc import ABC, abstractmethod
from functools import partial

from typing_extensions import Final

from praatio.utilities import errors
from praatio.utilities import utils

sampleWidthDict: Final = {1: "b", 2: "h", 4: "i", 8: "q"}

_KEEP: Final = "keep"
_DELETE: Final = "delete"

ZERO_CROSSING_TIMESTEP: Final = 0.002
DEFAULT_SINE_FREQUENCY = 200
NUM_BITS_IN_A_BYTE = 8


def calculateMaxAmplitude(sampleWidth: int) -> int:
    """Gets the largest possible amplitude representable by a given sample width

    The formula is 2^(n-1) - 1 where n is the number of bits
    - the first -1 is because the result is signed
    - the second -1 is because the value is 0 based
    e.g. if n=3 then 2^(3-1)-1 => 3
         if n=4 then 2^(4-1)-1 => 7

    Args:
        sampleWidth: the width in bytes of a sample in the wave file

    Returns:
        An integer
    """
    return 2 ** (sampleWidth * NUM_BITS_IN_A_BYTE - 1) - 1


def convertFromBytes(byteStr: bytes, sampleWidth: int) -> Tuple[int, ...]:
    """Convert frames of a python wave object from bytes to numbers"""
    byteCode = sampleWidthDict[sampleWidth]
    actualNumFrames = int(len(byteStr) / float(sampleWidth))
    audioFrameList = struct.unpack("<" + byteCode * actualNumFrames, byteStr)

    return audioFrameList


def convertToBytes(numList: Tuple[int, ...], sampleWidth: int) -> bytes:
    """Convert frames of a python wave object from numbers to bytes"""
    byteCode = sampleWidthDict[sampleWidth]
    byteStr = struct.pack("<" + byteCode * len(numList), *numList)

    return byteStr


def extractSubwav(fn: str, outputFN: str, startTime: float, endTime: float) -> None:
    """Get a subsegment of an audio file"""
    wav = QueryWav(fn)
    frames = wav.getFrames(startTime, endTime)
    wav.outputFrames(frames, outputFN)


def getDuration(fn: str) -> float:
    """Get the total duration of an audio file"""
    return QueryWav(fn).duration


def readFramesAtTime(
    audiofile: wave.Wave_read, startTime: float, endTime: float
) -> bytes:
    """Read the audio frames for the specified internal of an audio file"""
    params = audiofile.getparams()
    frameRate = params[2]

    audiofile.setpos(round(frameRate * startTime))
    frames = audiofile.readframes(round(frameRate * (endTime - startTime)))

    return frames


def readFramesAtTimes(
    audiofile: wave.Wave_read,
    keepIntervals: List[Tuple[float, float]] = None,
    deleteIntervals: List[Tuple[float, float]] = None,
    replaceFunc: Optional[Callable[[float], bytes]] = None,
) -> bytes:
    """Reads an audio file into memory, with some configuration

    Args:
        audiofile: the time to get the interval from
        keepIntervals: duration of the interval
        deleteIntervals: the maximum allowed time
        replaceFunc: is the interval before or after the targetTime?

    Returns:
        A bytestring of the loaded audio file

    Raises:
        ArgumentError: The timestamps in keepIntervals or deleteIntervals exceed the audio duration
        ArgumentError: Only one of keepIntervals and deleteIntervals can be specified
    """
    params = audiofile.getparams()
    frameRate = params[2]
    nframes = params[3]

    duration = nframes / float(frameRate)
    markedIntervals = _computeKeepDeleteIntervals(
        0.0, duration, keepIntervals, deleteIntervals
    )

    if markedIntervals[-1][1] > duration:
        raise errors.ArgumentError(
            "Timestamps in keepIntervals and deleteIntervals cannot exceed wav file duration"
        )

    # Grab the sections to be kept
    audioFrames: bytes = b""
    for start, end, label in markedIntervals:
        if label == _KEEP:
            audioFrames += readFramesAtTime(audiofile, start, end)

        # If we are not keeping a region and we're not shrinking the
        # duration, fill in the deleted portions with zeros
        elif label == _DELETE and replaceFunc:
            audioFrames += replaceFunc(end - start)

    return audioFrames


class AbstractWav(ABC):
    def __init__(self, params: List):
        self.params = params

        self.nchannels: int = params[0]
        self.sampleWidth: int = params[1]
        self.frameRate: int = params[2]
        self.nframes: int = params[3]
        self.comptype = params[4]
        self.compname = params[5]

        if self.nchannels != 1:
            raise (
                errors.ArgumentError(
                    "Only audio with a single channel can be loaded. Your file was #{self.nchannels}."
                )
            )

    def _iterZeroCrossings(
        self,
        start: float,
        withinThreshold,
        step: float,
        reverse: bool,
    ) -> Optional[float]:
        if not withinThreshold(start):
            return None

        startTime, endTime = utils.getInterval(start, step, self.duration, reverse)
        samples = self.getSamples(startTime, endTime)

        return _findNextZeroCrossing(startTime, samples, self.frameRate, reverse)

    @property
    @abstractmethod
    def duration(self) -> float:  # pragma: no cover
        pass

    def findNearestZeroCrossing(
        self, targetTime: float, timeStep: float = ZERO_CROSSING_TIMESTEP
    ) -> float:
        """Finds the nearest zero crossing at the given time in an audio file

        Looks both before and after the timeStamp
        """

        leftStartTime = rightStartTime = targetTime

        samplesPerStep = timeStep * self.frameRate
        if samplesPerStep < 2:
            raise errors.ArgumentError(
                f"'timeStep' ({timeStep}) must be large enough to contain "
                f"multiple samples for audio framerate ({self.frameRate})"
            )

        # Find zero crossings
        smallestLeft = None
        smallestRight = None
        oneSampleDuration = 1 / self.frameRate
        while True:
            # Increasing our timeStep by one sample enables
            # us to find zero-crossings that sit at the boundary
            # of two samples (two different iterations of this loop)
            smallestLeft = self._iterZeroCrossings(
                leftStartTime, lambda x: x > 0, timeStep + oneSampleDuration, True
            )
            smallestRight = self._iterZeroCrossings(
                rightStartTime,
                lambda x: x + timeStep < self.duration,
                timeStep + oneSampleDuration,
                False,
            )

            if smallestLeft is not None or smallestRight is not None:
                break
            # TODO: I think this case shouldn't be possible
            elif leftStartTime < 0 and rightStartTime > self.duration:
                raise (errors.FindZeroCrossingError(0, self.duration))
            else:
                # oneSampleDuration is not added here
                leftStartTime -= timeStep
                rightStartTime += timeStep

        # Under ordinary circumstances, this should not occur
        if smallestLeft is None and smallestRight is None:
            raise errors.FindZeroCrossingError(0, self.duration)

        return utils.chooseClosestTime(targetTime, smallestLeft, smallestRight)

    @abstractmethod
    def getFrames(self, startTime: float, endTime: float) -> bytes:  # pragma: no cover
        pass

    @abstractmethod
    def getSamples(
        self, startTime: float, endTime: float
    ) -> Tuple[int, ...]:  # pragma: no cover
        pass

    def outputFrames(self, frames: bytes, outputFN: str) -> None:
        """Output frames using the same parameters as this Wav"""
        outWave = wave.open(outputFN, "w")
        outWave.setparams(
            [
                self.nchannels,
                self.sampleWidth,
                self.frameRate,
                len(frames),
                self.comptype,
                self.compname,
            ]
        )
        outWave.writeframes(frames)


class QueryWav(AbstractWav):
    """A class for getting information about a wave file

    The wave file is never loaded--we only keep a reference to the
    file descriptor.  All operations on QueryWavs are fast.
    QueryWavs don't (shouldn't) change state.  For doing
    multiple modifications, use a Wav.
    """

    def __init__(self, fn: str):
        self.audiofile = wave.open(fn, "r")
        super(QueryWav, self).__init__(self.audiofile.getparams())

    @property
    def duration(self) -> float:
        duration = float(self.nframes) / self.frameRate
        return duration

    def getFrames(self, startTime: float = None, endTime: float = None) -> bytes:
        if startTime is None:
            startTime = 0

        if endTime is None:
            endTime = self.duration

        return readFramesAtTime(self.audiofile, startTime, endTime)

    def getSamples(self, startTime: float, endTime: float) -> Tuple[int, ...]:

        frames = self.getFrames(startTime, endTime)
        audioFrameList = convertFromBytes(frames, self.sampleWidth)

        return audioFrameList


class Wav(AbstractWav):
    """A class for manipulating audio files

    The wav file is represented by its wavform as a series of signed
    integers.  This can be very slow and take up lots of memory with
    large files.
    """

    def __init__(self, frames: bytes, params: List):
        self.frames = frames
        super(Wav, self).__init__(params)

    def __eq__(self, other):
        if not isinstance(other, Wav):
            return False

        return self.frames == other.frames

    def _getIndexAtTime(self, startTime: float) -> int:
        """Gets the index in the frame list for the given time"""
        return round(startTime * self.frameRate * self.sampleWidth)

    @classmethod
    def open(cls, fn: str) -> "Wav":
        wav = wave.open(fn, "r")
        audioFrames = readFramesAtTime(wav, startTime=0, endTime=getDuration(fn))
        return Wav(audioFrames, wav.getparams())

    def concatenate(self, frames: bytes) -> None:
        self.frames += frames

    def deleteSegment(self, startTime: float, endTime: float) -> None:
        i = self._getIndexAtTime(startTime)
        j = self._getIndexAtTime(endTime)
        self.frames = self.frames[:i] + self.frames[j:]

    @property
    def duration(self) -> float:
        return len(self.frames) / self.frameRate / self.sampleWidth

    def getFrames(self, startTime: float, endTime: float) -> bytes:
        i = self._getIndexAtTime(startTime)
        j = self._getIndexAtTime(endTime)
        return self.frames[i:j]

    def getSamples(self, startTime: float, endTime: float) -> Tuple[int, ...]:
        frames = self.getFrames(startTime, endTime)
        return convertFromBytes(frames, self.sampleWidth)

    def getSubwav(self, startTime: float, endTime: float) -> "Wav":
        frames = self.getFrames(startTime, endTime)
        return Wav(frames, self.params)

    def insert(self, startTime: float, frames: bytes) -> None:
        i = self._getIndexAtTime(startTime)
        self.frames = self.frames[:i] + frames + self.frames[i:]

    def new(self) -> "Wav":
        return copy.deepcopy(self)

    def replaceSegment(self, startTime: float, endTime: float, frames: bytes):
        self.deleteSegment(startTime, endTime)
        self.insert(startTime, frames)

    def save(self, outputFN: str) -> None:
        outWave = wave.open(outputFN, "w")
        outWave.setparams(
            [
                self.nchannels,
                self.sampleWidth,
                self.frameRate,
                len(self.frames),
                self.comptype,
                self.compname,
            ]
        )
        outWave.writeframes(self.frames)


class AudioGenerator:
    def __init__(self, sampleWidth, frameRate):
        self.sampleWidth: int = sampleWidth
        self.frameRate: int = frameRate

    @classmethod
    def fromWav(cls, wav: AbstractWav) -> "AudioGenerator":
        """Build an AudioGenerator with parameters derived from a Wav or QueryWav"""
        return AudioGenerator(wav.sampleWidth, wav.frameRate)

    def buildSineWaveGenerator(self, frequency, amplitude) -> Callable[[float], bytes]:
        """Returns a function that takes a duration and returns a generated sine wave"""
        return partial(
            self.generateSineWave,
            frequency=frequency,
            amplitude=amplitude,
        )

    def generateSineWave(
        self,
        duration: float,
        frequency: int,
        amplitude: Optional[float] = None,
    ) -> bytes:
        if amplitude is None:
            amplitude = calculateMaxAmplitude(self.sampleWidth)

        nSamples = round(duration * self.frameRate)
        wavSpec = 2 * math.pi * frequency / float(self.frameRate)
        sinWaveNums = [
            round(amplitude * math.sin(wavSpec * i)) for i in range(nSamples)
        ]
        return convertToBytes(tuple(sinWaveNums), self.sampleWidth)

    def generateSilence(self, duration: float) -> bytes:
        zeroBinValue = struct.pack(sampleWidthDict[self.sampleWidth], 0)
        return zeroBinValue * round(self.frameRate * duration)


def _findNextZeroCrossing(
    startTime: float,
    samples: Tuple[int, ...],
    frameRate: float,
    reverse: bool,
) -> Optional[float]:
    """Finds the nearest zero crossing, searching in one direction

    Can do a 'reverse' search by setting reverse to True.  In that case,
    the sample list is searched from back to front.

    targetTime is the startTime if reverse=False and
        the endTime if reverse=True
    """
    zeroI = _getNearestZero(samples, reverse)
    if zeroI is None:
        zeroI = _getZeroThresholdCrossing(samples, reverse)
        if zeroI is None:
            return None

    return startTime + zeroI / float(frameRate)


def _getNearestZero(samples: Tuple[int, ...], reverse: bool) -> Optional[int]:
    return utils.find(samples, 0, reverse)


def _getZeroThresholdCrossing(samples: Tuple[int, ...], reverse: bool) -> Optional[int]:
    signList = [utils.sign(val) for val in samples]
    changeList = [signList[i] != signList[i + 1] for i in range(len(samples) - 1)]
    zeroI = utils.find(changeList, True, reverse)

    if zeroI is None:
        return None

    # We found the zero by comparing points to the point adjacent to them.
    # It is possible the adjacent point is closer to zero than this one,
    # in which case, it is the better zeroedI
    if abs(samples[zeroI]) > abs(samples[zeroI + 1]):
        zeroI = zeroI + 1

    return zeroI


def _computeKeepDeleteIntervals(
    start: float,
    stop: float,
    keepIntervals: List[Tuple[float, float]] = None,
    deleteIntervals: List[Tuple[float, float]] = None,
) -> List[Tuple[float, float, str]]:
    """Returns a list of intervals, each one labeled 'keep' or 'delete'"""
    if keepIntervals and deleteIntervals:
        raise errors.ArgumentError(
            "You cannot specify both 'keepIntervals' or 'deleteIntervals'."
        )

    elif not keepIntervals and not deleteIntervals:
        computedKeepIntervals = [(start, stop)]
        computedDeleteIntervals = []

    elif deleteIntervals:
        deleteTimestamps = [(interval[0], interval[1]) for interval in deleteIntervals]
        computedKeepIntervals = utils.invertIntervalList(deleteTimestamps, start, stop)
        computedDeleteIntervals = deleteTimestamps

    elif keepIntervals:
        keepTimestamps = [(interval[0], interval[1]) for interval in keepIntervals]
        computedKeepIntervals = keepTimestamps
        computedDeleteIntervals = utils.invertIntervalList(keepTimestamps, start, stop)

    annotatedKeepIntervals = [
        (start, end, _KEEP) for start, end in computedKeepIntervals
    ]
    annotatedDeleteIntervals = [
        (start, end, _DELETE) for start, end in computedDeleteIntervals
    ]
    intervals = sorted(annotatedKeepIntervals + annotatedDeleteIntervals)

    return intervals



"""
An IntervalTier is a tier containing an array of intervals -- data that spans a period of time
"""
from typing import Callable, List, Optional, Tuple, Sequence

from typing_extensions import Literal


from praatio.utilities.constants import (
    Interval,
    INTERVAL_TIER,
    CropCollision,
)

from praatio.utilities import errors
from praatio.utilities import utils
from praatio.utilities import constants

from praatio.data_classes import textgrid_tier


def _homogenizeEntries(entries):
    """
    Enforces consistency in intervals

    - converts all entries to intervals
    - removes whitespace in labels
    - sorts values by time
    """
    processedEntries = [
        Interval(float(start), float(end), label.strip())
        for start, end, label in entries
    ]
    processedEntries.sort()
    return processedEntries


def _calculateMinAndMaxTime(entries: Sequence[Interval], minT=None, maxT=None):
    minTimeList = [interval.start for interval in entries]
    maxTimeList = [interval.end for interval in entries]

    if minT is not None:
        minTimeList.append(float(minT))
    if maxT is not None:
        maxTimeList.append(float(maxT))

    try:
        resolvedMinT = min(minTimeList)
        resolvedMaxT = max(maxTimeList)
    except ValueError:
        raise errors.TimelessTextgridTierException()

    return (resolvedMinT, resolvedMaxT)


class IntervalTier(textgrid_tier.TextgridTier):

    tierType = INTERVAL_TIER
    entryType = Interval

    def __init__(
        self,
        name: str,
        entries: List[Interval],
        minT: Optional[float] = None,
        maxT: Optional[float] = None,
    ):
        """An interval tier is for annotating events that have duration

        The entries is of the form:
        [(startTime1, endTime1, label1), (startTime2, endTime2, label2), ]

        The data stored in the labels can be anything but will
        be interpreted as text by praatio (the label could be descriptive
        text e.g. ('erase this region') or numerical data e.g. (average pitch
        values like '132'))
        """
        entries = _homogenizeEntries(entries)
        calculatedMinT, calculatedMaxT = _calculateMinAndMaxTime(entries, minT, maxT)

        super(IntervalTier, self).__init__(
            name, entries, calculatedMinT, calculatedMaxT
        )
        self._validate()

    def _validate(self):
        """An interval tier is invalid if the entries are out of order or overlap with each other"""
        for entry in self.entries:
            if entry.start >= entry.end:
                raise errors.TextgridStateError(
                    f"The start time of an interval ({entry.start}) "
                    f"cannot occur after its end time ({entry.end})"
                )

        for entry, nextEntry in zip(self.entries[0::], self.entries[1::]):
            if entry.end > nextEntry.start:
                raise errors.TextgridStateError(
                    "Two intervals in the same tier overlap in time:\n"
                    f"({entry.start}, {entry.end}, {entry.label}) and ({entry.start}, {entry.end}, {entry.label})"
                )

    def crop(
        self,
        cropStart: float,
        cropEnd: float,
        mode: Literal["strict", "lax", "truncated"],
        rebaseToZero: bool,
    ) -> "IntervalTier":
        """Creates a new tier with all entries that fit inside the new interval

        Args:
            cropStart:
            cropEnd:
            mode: determines cropping behavior
                - 'strict', only intervals wholly contained by the crop
                    interval will be kept
                - 'lax', partially contained intervals will be kept
                - 'truncated', partially contained intervals will be
                    truncated to fit within the crop region.
            rebaseToZero: if True, the cropped textgrid values
                will be subtracted by the cropStart

        Returns:
            the modified version of the current tier
        """

        utils.validateOption("mode", mode, CropCollision)

        if cropStart >= cropEnd:
            raise errors.ArgumentError(
                f"Crop error: start time ({cropStart}) must occur before end time ({cropEnd})"
            )

        newEntryList = utils.getIntervalsInInterval(
            cropStart, cropEnd, self.entries, mode
        )

        if rebaseToZero is True:
            newSmallestValue = newEntryList[0][0]
            if newSmallestValue < cropStart:
                timeDiff = newSmallestValue
            else:
                timeDiff = cropStart
            newEntryList = [
                Interval(start - timeDiff, end - timeDiff, label)
                for start, end, label in newEntryList
            ]
            minT = 0.0
            maxT = cropEnd - cropStart
        else:
            minT = cropStart
            maxT = cropEnd

        croppedTier = IntervalTier(self.name, newEntryList, minT, maxT)

        return croppedTier

    def deleteEntry(self, entry: Interval) -> None:
        """Removes an entry from the entries"""
        self._entries.pop(self._entries.index(entry))

    def difference(self, tier: "IntervalTier") -> "IntervalTier":
        """Takes the set difference of this tier and the given one

        Any overlapping portions of entries with entries in this textgrid
        will be removed from the returned tier.

        Args:
            tier: the tier to subtract from this one

        Returns:
            the modified version of the current tier
        """
        retTier = self.new()

        for entry in tier.entries:
            retTier = retTier.eraseRegion(
                entry.start,
                entry.end,
                collisionMode=constants.EraseCollision.TRUNCATE,
                doShrink=False,
            )

        return retTier

    def editTimestamps(
        self,
        offset: float,
        reportingMode: Literal["silence", "warning", "error"] = "warning",
    ) -> "IntervalTier":
        """Modifies all timestamps by a constant amount

        Args:
            offset: the amount to shift all intervals
            reportingMode: Determines the behavior if an entries moves outside
                of minTimestamp or maxTimestamp after being edited

        Returns:
            the modified version of the current tier
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )
        errorReporter = utils.getErrorReporter(reportingMode)

        newEntryList = []
        for interval in self.entries:

            newStart = offset + interval.start
            newEnd = offset + interval.end

            utils.checkIsUndershoot(newStart, self.minTimestamp, errorReporter)
            utils.checkIsOvershoot(newEnd, self.maxTimestamp, errorReporter)

            if newEnd <= 0:
                continue
            if newStart < 0:
                newStart = 0

            newEntryList.append(Interval(newStart, newEnd, interval.label))

        # Determine new min and max timestamps
        newMin = min([interval.start for interval in newEntryList])
        newMax = max([interval.end for interval in newEntryList])

        if newMin > self.minTimestamp:
            newMin = self.minTimestamp

        if newMax < self.maxTimestamp:
            newMax = self.maxTimestamp

        return IntervalTier(self.name, newEntryList, newMin, newMax)

    def eraseRegion(
        self,
        start: float,
        end: float,
        collisionMode: Literal["truncate", "categorical", "error"] = "error",
        doShrink: bool = True,
    ) -> "IntervalTier":
        """Makes a region in a tier blank (removes all contained entries)

        Args:
            start:
            end:
            collisionMode: Determines the behavior when the region to erase
                overlaps with existing intervals.
                - 'truncate' partially contained entries will have the portion
                    removed that overlaps with the target entry
                - 'categorical' all entries that overlap, even partially, with
                    the target entry will be completely removed
                - None or any other value throws IntervalCollision
            doShrink: If True, moves leftward by (/end/ - /start/)
                amount, each item that occurs after /end/

        Returns:
            The modified version of the current tier

        Raises:
            CollisionError
        """
        utils.validateOption("collisionMode", collisionMode, constants.EraseCollision)

        matchList = self.crop(start, end, CropCollision.LAX, False).entries
        newTier = self.new()

        if len(matchList) == 0:
            pass
        else:
            if collisionMode == constants.EraseCollision.ERROR:
                raise errors.CollisionError(
                    f"Erase region ({start}, {end})overlapped with an interval. "
                    "If this was expected, consider setting the collisionMode"
                )

            # Remove all the matches from the entries
            # Go in reverse order because we're destructively altering
            # the order of the list (messes up index order)
            for interval in matchList[::-1]:
                newTier.deleteEntry(interval)

            # If we're only truncating, reinsert entries on the left and
            # right edges
            # if categorical, it doesn't make it into the list at all
            if collisionMode == constants.EraseCollision.TRUNCATE:

                # Check left edge
                if matchList[0].start < start:
                    newEntry = Interval(matchList[0].start, start, matchList[0].label)
                    newTier.insertEntry(newEntry)

                # Check right edge
                if matchList[-1].end > end:
                    newEntry = Interval(end, matchList[-1].end, matchList[-1].label)
                    newTier.insertEntry(newEntry)

        if doShrink is True:

            diff = end - start
            newEntryList = []
            for interval in newTier.entries:
                if interval.end <= start:
                    newEntryList.append(interval)
                elif interval.start >= end:
                    newEntryList.append(
                        Interval(
                            interval.start - diff, interval.end - diff, interval.label
                        )
                    )

            # Special case: an interval that spanned the deleted
            # section
            for i in range(0, len(newEntryList) - 1):
                rightEdge = newEntryList[i].end == start
                leftEdge = newEntryList[i + 1].start == start
                sameLabel = newEntryList[i].label == newEntryList[i + 1].label
                if rightEdge and leftEdge and sameLabel:
                    newInterval = Interval(
                        newEntryList[i].start,
                        newEntryList[i + 1].end,
                        newEntryList[i].label,
                    )

                    newEntryList.pop(i + 1)
                    newEntryList.pop(i)
                    newEntryList.insert(i, newInterval)

                    # Only one interval can span the deleted section,
                    # so if we've found it, move on
                    break

            newMax = newTier.maxTimestamp - diff
            newTier = newTier.new(entries=newEntryList, maxTimestamp=newMax)

        return newTier

    def getValuesInIntervals(self, dataTupleList: List) -> List[Tuple[Interval, List]]:
        """Returns data from dataTupleList contained in labeled intervals

        Each labeled interval will get its own list of data values.

        dataTupleList should be of the form:
        [(time1, value1a, value1b,...), (time2, value2a, value2b...), ...]
        """

        returnList = []

        for interval in self.entries:
            intervalDataList = utils.getValuesInInterval(
                dataTupleList, interval.start, interval.end
            )
            returnList.append((interval, intervalDataList))

        return returnList

    def getNonEntries(self) -> List[Interval]:
        """Returns the regions of the textgrid without labels

        This can include unlabeled segments and regions marked as silent.
        """
        entries = self.entries
        invertedEntryList = [
            Interval(entries[i].end, entries[i + 1].start, "")
            for i in range(len(entries) - 1)
        ]

        # Remove entries that have no duration (ie lie between two entries
        # that share a border)
        invertedEntryList = [
            interval for interval in invertedEntryList if interval.start < interval.end
        ]

        if entries[0].start > 0:
            invertedEntryList.insert(0, Interval(0, entries[0].start, ""))

        if entries[-1].end < self.maxTimestamp:
            invertedEntryList.append(Interval(entries[-1].end, self.maxTimestamp, ""))

        invertedEntryList = [
            interval if isinstance(interval, Interval) else Interval(*interval)
            for interval in invertedEntryList
        ]

        return invertedEntryList

    def insertEntry(
        self,
        entry: Interval,
        collisionMode: Literal["replace", "merge", "error"] = "error",
        collisionReportingMode: Literal["silence", "warning"] = "warning",
    ) -> None:
        """Inserts an interval into the tier

        Args:
            entry: the Interval to insert
            collisionMode: determines the behavior in the event that intervals
                exist in the insertion area.
                - 'replace' will remove existing items
                - 'merge' will fuse the inserting item with existing items
                - None or any other value will throw a CollisionError
            collisionReportingMode: Determines the behavior if the new entry
                overlaps with an existing one

        Returns:
            the modified version of the current tier
        """
        utils.validateOption(
            "collisionMode", collisionMode, constants.IntervalCollision
        )
        utils.validateOption(
            "collisionReportingMode",
            collisionReportingMode,
            constants.ErrorReportingMode,
        )
        collisionReporter = utils.getErrorReporter(collisionReportingMode)

        if not isinstance(entry, Interval):
            interval = Interval(*entry)
        else:
            interval = entry

        matchList = self.crop(
            interval.start, interval.end, CropCollision.LAX, False
        )._entries

        if len(matchList) == 0:
            self._entries.append(interval)

        elif collisionMode == constants.IntervalCollision.REPLACE:
            for matchEntry in matchList:
                self.deleteEntry(matchEntry)
            self._entries.append(interval)

        elif collisionMode == constants.IntervalCollision.MERGE:
            for matchEntry in matchList:
                self.deleteEntry(matchEntry)
            matchList.append(interval)
            matchList.sort()  # By starting time

            newInterval = Interval(
                min([tmpInterval.start for tmpInterval in matchList]),
                max([tmpInterval.end for tmpInterval in matchList]),
                "-".join([tmpInterval.label for tmpInterval in matchList]),
            )
            self._entries.append(newInterval)

        else:
            raise errors.CollisionError(
                "Attempted to insert interval "
                f"({interval.start}, {interval.end}, '{interval.label}') into tier {self.name} "
                "of textgrid but overlapping entries "
                f"{[tuple(interval) for interval in matchList]} "
                "already exist"
            )

        self.sort()

        if self._entries[0][0] < self.minTimestamp:
            self.minTimestamp = self._entries[0][0]

        if self._entries[-1][1] > self.maxTimestamp:
            self.maxTimestamp = self._entries[-1][1]

        if len(matchList) != 0:
            collisionReporter(
                errors.CollisionError,
                f"Collision warning for ({interval}) with items "
                f"({matchList}) of tier '{self.name}'",
            )

    def insertSpace(
        self,
        start: float,
        duration: float,
        collisionMode: Literal["stretch", "split", "no_change", "error"],
    ) -> "IntervalTier":
        """Inserts a blank region into the tier

        Args:
            start:
            duration:
            collisionMode: Determines the behavior that occurs if
                an interval stradles the starting point
                - 'stretch' stretches the interval by /duration/ amount
                - 'split' splits the interval into two--everything to the
                    right of 'start' will be advanced by 'duration' seconds
                - 'no change' leaves the interval as is with no change
                - 'error' will stop execution and raise an error

        Returns:
            the modified version of the current tier
        """
        utils.validateOption(
            "collisionMode", collisionMode, constants.WhitespaceCollision
        )

        newEntryList = []
        for interval in self.entries:
            # Entry exists before the insertion point
            if interval.end <= start:
                newEntryList.append(interval)
            # Entry exists after the insertion point
            elif interval.start >= start:
                newEntryList.append(
                    Interval(
                        interval.start + duration,
                        interval.end + duration,
                        interval.label,
                    )
                )
            # Entry straddles the insertion point
            elif interval.start <= start and interval.end > start:
                if collisionMode == constants.WhitespaceCollision.STRETCH:
                    newEntryList.append(
                        Interval(
                            interval.start, interval.end + duration, interval.label
                        )
                    )
                elif collisionMode == constants.WhitespaceCollision.SPLIT:
                    # Left side of the split
                    newEntryList.append(Interval(interval.start, start, interval.label))
                    # Right side of the split
                    newEntryList.append(
                        (
                            start + duration,
                            start + duration + (interval.end - start),
                            interval.label,
                        )
                    )
                elif collisionMode == constants.WhitespaceCollision.NO_CHANGE:
                    newEntryList.append(interval)
                else:
                    raise errors.ArgumentError(
                        f"Collision occured during insertSpace() for interval '{interval}' "
                        f"and given white space insertion interval ({start}, {start + duration})"
                    )

        newTier = self.new(
            entries=newEntryList, maxTimestamp=self.maxTimestamp + duration
        )

        return newTier

    def intersection(self, tier: "IntervalTier", demarcator="-") -> "IntervalTier":
        """Takes the set intersection of this tier and the given one

        - The output will contain one interval for each overlapping pair
          e.g. [(1, 2, 'foo')] and [(1, 1.3, 'bang'), (1.7, 2, 'wizz')]
                -> [(1, 1.3, 'foo-bang'), (1.7, 2, 'foo-wizz')]
        - Only intervals that exist in both tiers will remain in the returned tier.
          e.g. [(1, 2, 'foo'), (3, 4, 'bar')] and [(1, 2, 'bang'), (2, 3, 'wizz')]
                -> [(1, 2, 'foo-bang')]
        - If intervals partially overlap, only the overlapping portion will be returned.
          e.g. [(1, 2, 'foo')] and [(0.5, 1.5, 'bang')]
                -> [(1, 1.5, 'foo-bang')]

        Compare with IntervalTier.mergeLabels

        Args:
            tier: the tier to intersect with
            demarcator: the character to separate the labels of the overlapping intervals

        Returns:
            IntervalTier: the modified version of the current tier
        """
        retEntryList = []
        for interval in tier.entries:
            subTier = self.crop(
                interval.start, interval.end, CropCollision.TRUNCATED, False
            )

            # Combine the labels in the two tiers
            subEntryList = [
                (
                    subInterval.start,
                    subInterval.end,
                    f"{subInterval.label}{demarcator}{interval.label}",
                )
                for subInterval in subTier.entries
            ]

            retEntryList.extend(subEntryList)

        newName = f"{self.name}-{tier.name}"

        retTier = self.new(newName, retEntryList)

        return retTier

    def mergeLabels(
        self, tier: "IntervalTier", demarcator: str = ","
    ) -> "IntervalTier":
        """Merges labels of overlapping tiers into this tier

        - All intervals in this tier will appear in the output; for the given tier, only intervals
          that overlap with content in this tier will appear in the output
          e.g. [(1, 2, 'foo'), (3, 4, 'bar')] and [(1, 2, 'bang'), (2, 3, 'wizz')]
                -> [(1, 2, 'foo(bang)'), (3, 4, 'bar()')]
        - If multiple entries exist in a subinterval, their labels will be concatenated
          e.g. [(1, 2, 'hi')] and [(1, 1.5, 'h'), (1.5, 2, 'ai')] -> [(1, 2, 'hi(h,ai)')]

        compare with IntervalTier.intersection

        Args:
            tier: the tier to intersect with
            demarcator: the string to separate items that fall in the same subinterval

        Returns:
            IntervalTier: the modified version of the current tier
        """
        retEntryList = []
        for interval in self.entries:
            subTier = tier.crop(
                interval.start, interval.end, CropCollision.TRUNCATED, False
            )
            if len(subTier._entries) == 0:
                continue

            subLabel = demarcator.join([entry.label for entry in subTier.entries])
            label = f"{interval.label}({subLabel})"

            start = min(interval.start, subTier._entries[0].start)
            end = max(interval.end, subTier._entries[-1].end)

            intersectedInterval = (
                start,
                end,
                label,
            )

            retEntryList.append(intersectedInterval)

        newName = f"{self.name}-{tier.name}"

        retTier = self.new(newName, retEntryList)

        return retTier

    def morph(
        self,
        targetTier: "IntervalTier",
        filterFunc: Optional[Callable[[str], bool]] = None,
    ) -> "IntervalTier":
        """Morphs the duration of segments in this tier to those in another

        This preserves the labels and the duration of silence in
        this tier while changing the duration of labeled segments.

        Args:
            targetTier:
            filterFunc: if specified, filters entries. The
                functor takes one argument, an Interval. It returns true
                if the Interval should be modified and false if not.

        Returns:
            The modified version of the current tier
        """
        cumulativeAdjustAmount = 0
        newEntryList = []
        allIntervals = [self.entries, targetTier.entries]
        for sourceInterval, targetInterval in utils.safeZip(allIntervals, True):

            # sourceInterval.start - lastFromEnd -> was this interval and the
            # last one adjacent?
            newStart = sourceInterval.start + cumulativeAdjustAmount

            currIntervalDuration = sourceInterval.end - sourceInterval.start
            if filterFunc is None or filterFunc(sourceInterval.label):
                newIntervalDuration = targetInterval.end - targetInterval.start
                cumulativeAdjustAmount += newIntervalDuration - currIntervalDuration
                newEnd = newStart + newIntervalDuration
            else:
                newEnd = newStart + currIntervalDuration

            newEntryList.append(Interval(newStart, newEnd, sourceInterval.label))

        newMin = self.minTimestamp
        cumulativeDifference = newEntryList[-1].end - self.entries[-1].end
        newMax = self.maxTimestamp + cumulativeDifference

        return IntervalTier(self.name, newEntryList, newMin, newMax)

    def validate(
        self, reportingMode: Literal["silence", "warning", "error"] = "warning"
    ) -> bool:
        """Validate this tier

        Args:
            reportingMode (str): Determines the behavior if validation fails.

        Returns:
            True if the tier is valid; False if not
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )
        errorReporter = utils.getErrorReporter(reportingMode)

        isValid = True
        previousInterval = None
        for interval in self.entries:
            if interval.start >= interval.end:
                isValid = False
                errorReporter(
                    errors.TextgridStateError,
                    f"Invalid interval. End time occurs before or on the start time({interval}).",
                )

            if previousInterval and previousInterval.end > interval.start:
                isValid = False
                errorReporter(
                    errors.TextgridStateError,
                    f"Intervals are not sorted in time: "
                    f"[({previousInterval}), ({interval})]",
                )

            if utils.checkIsUndershoot(
                interval.start, self.minTimestamp, errorReporter
            ):
                isValid = False

            if utils.checkIsOvershoot(interval.end, self.maxTimestamp, errorReporter):
                isValid = False

            previousInterval = interval

        return isValid



"""Functions for reading/writing/manipulating klattgrid files

A klattgrid can be used for speech synthesis/resynthesis.
For more information on the praat klattgrid:
http://www.fon.hum.uva.nl/praat/manual/KlattGrid.html

There are three kinds of data types in a klattgrid:
null tiers (contain no data points -- seem to function as headers for a
            set of regular tiers)
regular tiers
embedded tiers

In this code:
null tiers and regular tiers are both represented by KlattPointTier

embedded tiers contain tiers of tiers (3 layers)
A KlattContainerTier contains a list of KlattIntermediateTiers which
contains a list of KlattSubPointTiers.  Only the KlattSubPointTiers contain
any points

see **examples/klatt_resynthesis.py**
"""

import io
from os.path import join
from typing import List, Tuple, Optional

from praatio.data_classes.klattgrid import (
    Klattgrid,
    KlattPointTier,
    KlattContainerTier,
    KlattSubPointTier,
    KlattIntermediateTier,
)
from praatio.utilities import utils


def openKlattgrid(fnFullPath: str) -> Klattgrid:

    try:
        with io.open(fnFullPath, "r", encoding="utf-16") as fd:
            data = fd.read()
    except UnicodeError:
        with io.open(fnFullPath, "r", encoding="utf-8") as fd:
            data = fd.read()
    data = data.replace("\r\n", "\n")

    # Right now, can only open normal klatt grid and not short ones
    kg = _openNormalKlattgrid(data)

    return kg


def wavToKlattgrid(
    praatEXE: str,
    inputFullPath: str,
    outputFullPath: str,
    timeStep: float = 0.005,
    numFormants: int = 5,
    maxFormantFreq: float = 5500.0,
    windowLength: float = 0.025,
    preEmphasis: float = 50.0,
    pitchFloor: float = 60.0,
    pitchCeiling: float = 600.0,
    minPitch: float = 50.0,
    subtractMean: bool = True,
    scriptFN: Optional[str] = None,
) -> None:
    """Extracts the klattgrid from a wav file

    The default values are the same as the ones used in praat
    """

    if subtractMean is True:
        subtractMeanStr = "yes"
    else:
        subtractMeanStr = "no"

    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "sound_to_klattgrid.praat")

    utils.runPraatScript(
        praatEXE,
        scriptFN,
        [
            inputFullPath,
            outputFullPath,
            timeStep,
            numFormants,
            maxFormantFreq,
            windowLength,
            preEmphasis,
            pitchFloor,
            pitchCeiling,
            minPitch,
            subtractMeanStr,
        ],
    )


def resynthesize(
    praatEXE: str,
    wavFN: str,
    klattFN: str,
    outputWavFN: str,
    doCascade: bool = True,
    scriptFN: Optional[str] = None,
) -> None:

    if doCascade:
        method = "Cascade"
    else:
        method = "Parallel"

    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "resynthesize_from_klattgrid.praat")

    #  Praat crashes on exit after resynthesis with a klattgrid
    utils.runPraatScript(praatEXE, scriptFN, [wavFN, klattFN, outputWavFN, method])


def _openNormalKlattgrid(data: str) -> Klattgrid:

    kg = Klattgrid()

    # Toss header
    data = data.split("\n\n", 1)[1]

    # Not sure if this is needed
    startI = data.index("points")
    startI = data.index("\n", startI)

    # Find sections
    sectionIndexList = _findIndicies(data, "<exists>")

    sectionIndexList.append(-1)

    for i in range(len(sectionIndexList) - 1):
        dataTuple = _getSectionHeader(data, sectionIndexList, i)
        name, minT, maxT, sectionData, sectionTuple = dataTuple

        # Container Tier
        if name in [
            "oral_formants",
            "nasal_formants",
            "nasal_antiformants",
            "tracheal_formants",
            "tracheal_antiformants",
            "delta_formants",
            "frication_formants",
        ]:

            kct = _proccessContainerTierInput(sectionData, name)
            kg.addTier(kct)

        else:

            # Process entries if this tier has any
            entries = _buildEntries(sectionTuple)
            tier = KlattPointTier(name, entries, minT, maxT)
            kg.addTier(tier)

    return kg


def _proccessContainerTierInput(sectionData: str, name: str):
    sectionData = sectionData.split("\n", 3)[-1]

    formantIndexList = _findIndicies(sectionData, "formants")

    subFilterList = [
        "bandwidths",
        "oral_formants_amplitudes",
        "nasal_formants_amplitudes",
        "tracheal_formants_amplitudes",
        "frication_formants_amplitudes",
    ]

    # Find the index of all the different data sections
    subFilterIndexList = [
        _findIndicies(sectionData, subName) for subName in subFilterList
    ]

    # 'Formant' search query finds duplicates -- remove them
    newFormantList = []
    for value in formantIndexList:
        if all([value not in subList for subList in subFilterIndexList]):
            newFormantList.append(value)
    formantIndexList = newFormantList

    # Combine regular query with formant query
    indexListOfLists = [
        formantIndexList,
    ] + subFilterIndexList

    # Flatten index list
    masterIndexList = [value for sublist in indexListOfLists for value in sublist]
    masterIndexList.sort()

    # If an index list is last, it it needs to include '-1' to capture the
    # rest of the data
    for subList in indexListOfLists:
        try:
            val = subList[-1]
        except IndexError:
            continue
        ii = masterIndexList.index(val)  # Index of the index
        try:
            subList.append(masterIndexList[ii + 1] - 1)
        except IndexError:
            subList.append(-1)

    # Build the tier structure
    kct = KlattContainerTier(name)
    for indexList in indexListOfLists:
        if indexList == []:
            continue
        tierList = []
        for j in range(len(indexList) - 1):
            try:
                tmpTuple = _getSectionHeader(sectionData, indexList, j)
            except ValueError:
                continue
            subName, subMin, subMax, _, subTuple = tmpTuple
            subName = subName[:-1]

            entries = _buildEntries(subTuple)
            tier = KlattSubPointTier(subName, entries, subMin, subMax)
            tierList.append(tier)
        kit = KlattIntermediateTier(subName.split()[0])
        for tier in tierList:
            kit.addTier(tier)
        kct.addTier(kit)

    return kct


def _findIndicies(data, keyword):
    indexList = utils.findAll(data, keyword)
    indexList = [data.rfind("\n", 0, i) for i in indexList]

    return indexList


def _getSectionHeader(data, indexList, i):
    sectionStart = indexList[i]
    sectionEnd = indexList[i + 1]
    sectionData = data[sectionStart:sectionEnd].strip()
    sectionTuple = sectionData.split("\n", 4)

    subheader, minr, maxr = sectionTuple[:3]
    name = subheader.split("?")[0].strip()
    minT = float(minr.split("=")[1].strip())
    maxT = float(maxr.split("=")[1].strip())

    tail = sectionTuple[3:]

    return name, minT, maxT, sectionData, tail


def _buildEntries(sectionTuple):
    entries = []
    if len(sectionTuple) > 1:  # Has points
        npoints = float(sectionTuple[0].split("=")[1].strip())
        if npoints > 0:
            entries = _processSectionData(sectionTuple[1])

    return entries


def _processSectionData(sectionData: str) -> List[Tuple[float, float]]:

    sectionData += "\n"

    startI = 0
    tupleList = []
    while True:
        try:
            startI = sectionData.index("=", startI) + 1  # Past the equal sign
        except ValueError:  # No more data
            break

        endI = sectionData.index("\n", startI)
        time = float(sectionData[startI:endI].strip())

        startI = sectionData.index("=", endI) + 1  # Just past the '=' sign
        endI = sectionData.index("\n", startI)
        value = float(sectionData[startI:endI].strip())

        startI = endI
        tupleList.append((time, value))

    return tupleList



"""
Python wrappers for various praat scripts contained in /praatScripts.

see **examples/auto_segment_speech.py**, **examples/get_pitch_and_formants.py**,
**klatt_resynthesis.py**
"""

import os
from os.path import join
import io
import csv
from typing import List, Optional, Tuple

from praatio import audio
from praatio import data_points
from praatio.utilities import constants
from praatio.utilities import utils

SILENCE_LABEL = "silent"
SOUND_LABEL = "sound"


def changeGender(
    praatEXE: str,
    wavFN: str,
    outputWavFN: str,
    pitchFloor: float,
    pitchCeiling: float,
    formantShiftRatio: float,
    pitchMedian: float = 0.0,
    pitchRange: float = 1.0,
    duration: float = 1.0,
    scriptFN: Optional[str] = None,
) -> None:
    """Changes the speech formants in a file using praat's change gender function

    PitchMedian = 0.0; no change in median pitch
    PitchRange = 1.0; no change in pitch range

    Uses the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__Change_gender___.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "change_gender.praat")

    #  Praat crashes on exit after resynthesis with a klattgrid
    utils.runPraatScript(
        praatEXE,
        scriptFN,
        [
            wavFN,
            outputWavFN,
            pitchFloor,
            pitchCeiling,
            formantShiftRatio,
            pitchMedian,
            pitchRange,
            duration,
        ],
    )


def changeIntensity(
    praatEXE: str,
    wavFN: str,
    outputWavFN: str,
    newIntensity: float,
    scriptFN: Optional[str] = None,
) -> None:
    """Changes the intensity of the wavFN (in db)

    Uses the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__Scale_intensity___.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "change_intensity.praat")

    #  Praat crashes on exit after resynthesis with a klattgrid
    utils.runPraatScript(praatEXE, scriptFN, [wavFN, outputWavFN, newIntensity])


def getFormants(
    praatEXE: str,
    inputWavFN: str,
    outputTxtFN: str,
    maxFormant: float,
    stepSize: float = 0.01,
    window_length: float = 0.025,
    preemphasis: float = 50,
    scriptFN: Optional[str] = None,
    undefinedValue: Optional[str] = None,
) -> List:
    """Get F1, F2, and F3 for the audio file

    maxFormant = 5500 for females, 5000 for males, <8000 for children

    Uses the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_Formant__burg____.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "get_formants.praat")

    argList = [
        inputWavFN,
        outputTxtFN,
        stepSize,
        maxFormant,
        window_length,
        preemphasis,
        -1,
        -1,
    ]
    utils.runPraatScript(praatEXE, scriptFN, argList)

    # Load the output
    path, fn = os.path.split(outputTxtFN)
    with open(os.path.join(path, fn), "r", encoding="utf-8") as fd:
        csvFile = csv.reader(fd)
        dataList = [row for row in csvFile]

    # The new praat script includes a header
    if dataList[0][0] == "time":
        dataList = dataList[1:]

    # Handle undefined values, convert values to float
    returnList = []
    for row in dataList:
        keep = True
        for i in range(1, 4):
            if "--" in row[i]:
                if undefinedValue is not None:
                    row[i] = undefinedValue
                else:
                    keep = False
                    break

        if keep is True:
            returnList.append([float(val) for val in row])

    return returnList


def getPulses(
    praatEXE: str,
    inputWavFN: str,
    outputPointTierFN: str,
    minPitch: float,
    maxPitch: float,
    scriptFN: Optional[str] = None,
) -> data_points.PointObject1D:
    """Gets the pitch/glottal pulses for an audio file.

    Uses the following praat command:
    http://www.fon.hum.uva.nl/praat/manual/Sound___Pitch__To_PointProcess__peaks____.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "get_pulses.praat")

    argList = [inputWavFN, outputPointTierFN, minPitch, maxPitch]
    utils.runPraatScript(praatEXE, scriptFN, argList)

    # Load the output
    pointObj = data_points.open1DPointObject(outputPointTierFN)

    return pointObj


def getSpectralInfo(
    praatEXE: str,
    inputWavFN: str,
    inputTGFN: str,
    outputCSVFN: str,
    tierName: str,
    spectralPower: int = 2,
    spectralMoment: int = 3,
    scriptFN: Optional[str] = None,
) -> Tuple[List, List]:
    """Extracts various spectral measures from an audio file

    Measures include: center_of_gravity, standard_deviation
    skewness, kertosis, central_movement

    Uses the following praat command:
    http://www.fon.hum.uva.nl/praat/manual/Spectrum.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "get_spectral_info.praat")

    argList = [
        inputWavFN,
        inputTGFN,
        outputCSVFN,
        tierName,
        spectralPower,
        spectralMoment,
    ]
    utils.runPraatScript(praatEXE, scriptFN, argList)

    # Load the output
    with io.open(outputCSVFN, "r", encoding="utf-8") as fd:
        data = fd.read()

    dataList = data.rstrip().split("\n")
    dataListOfLists = [row.split(",") for row in dataList]
    titleRow, mainDataListOfLists = dataListOfLists[0], dataListOfLists[1:]

    return titleRow, mainDataListOfLists


def resynthesizePitch(
    praatEXE: str,
    inputWavFN: str,
    pitchFN: str,
    outputWavFN: str,
    minPitch: float,
    maxPitch: float,
    scriptFN: Optional[str] = None,
    pointList: Optional[List] = None,
) -> None:
    """Resynthesizes the pitch in a wav file with the given pitch contour file

    The pitch track to use can optionally be passed in as pointList.  If
    so, it will be saved as pitchFN for praat to be able to use.

    Uses the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Manipulation.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "resynthesize_pitch.praat")

    if pointList is not None:
        dur = audio.QueryWav(inputWavFN).duration
        pointObj = data_points.PointObject2D(
            pointList, constants.DataPointTypes.PITCH, 0, dur
        )
        pointObj.save(pitchFN)

    utils.runPraatScript(
        praatEXE, scriptFN, [inputWavFN, pitchFN, outputWavFN, minPitch, maxPitch]
    )


def resynthesizeDuration(
    praatEXE: str,
    inputWavFN: str,
    durationTierFN: str,
    outputWavFN: str,
    minPitch: float,
    maxPitch: float,
    scriptFN: Optional[str] = None,
) -> None:
    """Resynthesizes the duration in a wav file with the given duration tier

    Uses the following praat command:
    https://www.fon.hum.uva.nl/praat/manual/Manipulation.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "resynthesize_duration.praat")

    utils.runPraatScript(
        praatEXE,
        scriptFN,
        [inputWavFN, durationTierFN, outputWavFN, minPitch, maxPitch],
    )


def annotateSilences(
    praatEXE: str,
    inputWavPath: str,
    outputTGPath: str,
    minPitch: float = 100,
    timeStep: float = 0.0,
    silenceThreshold: float = -25.0,
    minSilDur: float = 0.1,
    minSoundDur: float = 0.1,
    silentLabel: str = SILENCE_LABEL,
    soundLabel: str = SOUND_LABEL,
    scriptFN: Optional[str] = None,
) -> None:
    """Marks the silences and non-silences of an audio file

    Uses the praat command:
    https://www.fon.hum.uva.nl/praat/manual/Sound__To_TextGrid__silences____.html
    """
    if scriptFN is None:
        scriptFN = join(utils.scriptsPath, "annotate_silences.praat")

    utils.runPraatScript(
        praatEXE,
        scriptFN,
        [
            inputWavPath,
            outputTGPath,
            minPitch,
            timeStep,
            silenceThreshold,
            minSilDur,
            minSoundDur,
            silentLabel,
            soundLabel,
        ],
    )



"""
Functions for reading/writing/manipulating Textgrid classes

This is the 'heart' of praatio.
"""
import io
import copy
from typing import Optional, Tuple, Sequence
from typing_extensions import Literal
from collections import OrderedDict


from praatio.utilities.constants import (
    TextgridFormats,
    MIN_INTERVAL_LENGTH,
    CropCollision,
)

from praatio.data_classes import textgrid_tier
from praatio.data_classes import point_tier
from praatio.data_classes import interval_tier
from praatio.utilities import constants
from praatio.utilities import errors
from praatio.utilities import my_math
from praatio.utilities import textgrid_io
from praatio.utilities import utils


class Textgrid:
    """A container that stores and operates over interval and point tiers

    Textgrids are used by the Praat software to group tiers.  Each tier
    contains different annotation information for an audio recording.

    Attributes:
        tierNames(Tuple[str]): the list of tier names in the textgrid
        tiers(Tuple[TextgridTier]): the list of ordered tiers
        minTimestamp(float): the minimum allowable timestamp in the textgrid
        maxTimestamp(float): the maximum allowable timestamp in the textgrid
    """

    def __init__(self, minTimestamp: float = None, maxTimestamp: float = None):
        """Constructor for Textgrids

        Args:
            minTimestamp: the minimum allowable timestamp in the textgrid
            maxTimestamp: the maximum allowable timestamp in the textgrid
        """

        self._tierDict: OrderedDict[str, textgrid_tier.TextgridTier] = OrderedDict()

        # Timestamps are determined by the first tier added
        self.minTimestamp: float = minTimestamp  # type: ignore[assignment]
        self.maxTimestamp: float = maxTimestamp  # type: ignore[assignment]

    def __eq__(self, other):
        if not isinstance(other, Textgrid):
            return False

        isEqual = True
        isEqual &= my_math.isclose(self.minTimestamp, other.minTimestamp)
        isEqual &= my_math.isclose(self.maxTimestamp, other.maxTimestamp)

        isEqual &= self.tierNames == other.tierNames
        if isEqual:
            for tierName in self.tierNames:
                isEqual &= self.getTier(tierName) == other.getTier(tierName)

        return isEqual

    @property
    def tierNames(self) -> Tuple[str, ...]:
        return tuple(self._tierDict.keys())

    @property
    def tiers(self) -> Tuple[textgrid_tier.TextgridTier, ...]:
        return tuple(self._tierDict.values())

    def addTier(
        self,
        tier: textgrid_tier.TextgridTier,
        tierIndex: Optional[int] = None,
        reportingMode: Literal["silence", "warning", "error"] = "warning",
    ) -> None:
        """Add a tier to this textgrid.

        Args:
            tier: The tier to add to the textgrid
            tierIndex: Insert the tier into the specified position;
                if None, the tier will appear after all existing tiers
            reportingMode: This flag determines the behavior if there is a size
                difference between the maxTimestamp in the tier and the current
                textgrid.

        Returns:
            None

        Raises:
            TierNameExistsError: The textgrid already contains a tier with the same
                name as the tier being added
            TextgridStateAutoModified: The minimum or maximum timestamp was changed
                when not permitted
            IndexError: TierIndex is too large for the size of the existing tier list
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )
        errorReporter = utils.getErrorReporter(reportingMode)

        if tier.name in self.tierNames:
            raise errors.TierNameExistsError("Tier name already in tier")

        if tierIndex is None:
            self._tierDict[tier.name] = tier
        else:  # Need to recreate the tierDict with the new order
            newOrderedTierNameList = list(self.tierNames)
            newOrderedTierNameList.insert(tierIndex, tier.name)

            newTierDict = OrderedDict()
            self._tierDict[tier.name] = tier
            for tmpName in newOrderedTierNameList:
                newTierDict[tmpName] = self.getTier(tmpName)
            self._tierDict = newTierDict

        minV = tier.minTimestamp
        if self.minTimestamp is not None and minV < self.minTimestamp:
            errorReporter(
                errors.TextgridStateAutoModified,
                f"Minimum timestamp in Textgrid changed from ({self.minTimestamp}) to ({minV})",
            )
        if self.minTimestamp is None or minV < self.minTimestamp:
            self.minTimestamp = minV

        maxV = tier.maxTimestamp
        if self.maxTimestamp is not None and maxV > self.maxTimestamp:
            errorReporter(
                errors.TextgridStateAutoModified,
                f"Maximum timestamp in Textgrid changed from ({self.maxTimestamp}) to ({maxV})",
            )
        if self.maxTimestamp is None or maxV > self.maxTimestamp:
            self.maxTimestamp = maxV

    def appendTextgrid(self, tg: "Textgrid", onlyMatchingNames: bool) -> "Textgrid":
        """Append one textgrid to the end of this one

        Args:
            tg: the textgrid to add to this one
            onlyMatchingNames: if False, tiers that don't appear in both
                textgrids will also appear

        Returns:
            the modified version of the current textgrid
        """
        minTime = self.minTimestamp
        maxTime = self.maxTimestamp + tg.maxTimestamp
        retTG = Textgrid(minTime, maxTime)

        # Get all tier names.  Ordered first by this textgrid and
        # then by the other textgrid.
        combinedTierNames = list(self.tierNames)
        for tierName in tg.tierNames:
            if tierName not in combinedTierNames:
                combinedTierNames.append(tierName)

        # Determine the tier names that will be in the final textgrid
        finalTierNames = []
        if onlyMatchingNames is False:
            finalTierNames = combinedTierNames
        else:
            for tierName in combinedTierNames:
                if tierName in self.tierNames:
                    if tierName in tg.tierNames:
                        finalTierNames.append(tierName)

        # Add tiers from this textgrid
        for tierName in finalTierNames:
            if tierName in self.tierNames:
                tier = self.getTier(tierName)
                retTG.addTier(tier)

        # Add tiers from the given textgrid
        for tierName in finalTierNames:
            if tierName in tg.tierNames:
                appendTier = tg.getTier(tierName)
                appendTier = appendTier.new(minTimestamp=minTime, maxTimestamp=maxTime)

                appendTier = appendTier.editTimestamps(self.maxTimestamp)

                if tierName in retTG.tierNames:
                    tier = retTG.getTier(tierName)
                    newEntries = retTG.getTier(tierName).entries
                    newEntries += appendTier.entries

                    tier = tier.new(
                        entries=newEntries,
                        minTimestamp=minTime,
                        maxTimestamp=maxTime,
                    )
                    retTG.replaceTier(tierName, tier)

                else:
                    tier = appendTier
                    tier = tier.new(minTimestamp=minTime, maxTimestamp=maxTime)
                    retTG.addTier(tier)

        return retTG

    def crop(
        self,
        cropStart: float,
        cropEnd: float,
        mode: Literal["strict", "lax", "truncated"],
        rebaseToZero: bool,
    ) -> "Textgrid":
        """Creates a textgrid where all intervals fit within the crop region

        Args:
            cropStart: The start time of the crop interval
            cropEnd: The stop time of the crop interval
            mode: Determines the crop behavior
                - 'strict', only intervals wholly contained by the crop
                    interval will be kept
                - 'lax', partially contained intervals will be kept
                - 'truncated', partially contained intervals will be
                    truncated to fit within the crop region.
            rebaseToZero: if True, the cropped textgrid timestamps will be
                subtracted by the cropStart; if False, timestamps will not
                be changed

        Returns:
            the modified version of the current textgrid
        """
        utils.validateOption("mode", mode, CropCollision)

        if cropStart >= cropEnd:
            raise errors.ArgumentError(
                f"Crop error: start time ({cropStart}) must occur before end time ({cropEnd})"
            )

        if rebaseToZero is True:
            minT = 0.0
            maxT = cropEnd - cropStart
        else:
            minT = cropStart
            maxT = cropEnd
        newTG = Textgrid(minT, maxT)

        for tierName in self.tierNames:
            tier = self.getTier(tierName)
            newTier = tier.crop(cropStart, cropEnd, mode, rebaseToZero)

            reportingMode: Literal[
                "silence", "warning", "error"
            ] = constants.ErrorReportingMode.WARNING
            if mode == constants.CropCollision.LAX:
                # We expect that there will be changes to the size
                # of the textgrid when the mode is LAX
                reportingMode = constants.ErrorReportingMode.SILENCE

            newTG.addTier(newTier, reportingMode=reportingMode)

        return newTG

    def eraseRegion(self, start: float, end: float, doShrink: bool) -> "Textgrid":
        """Makes a region in a tier blank (removes all contained entries)

        Intervals that span the region to erase will be truncated.

        Args:
            start:
            end:
            doShrink: if True, all entries appearing after the
                erased interval will be shifted to fill the void (ie
                the duration of the textgrid will be reduced by
                *start* - *end*)

        Returns:
            the modified version of the current textgrid

        Raises:
            ArgumentError
        """
        if start >= end:
            raise errors.ArgumentError(
                f"EraseRegion error: start time ({start}) must occur before end time ({end})"
            )

        diff = end - start

        maxTimestamp = self.maxTimestamp
        if doShrink is True:
            maxTimestamp -= diff

        newTG = Textgrid(self.minTimestamp, self.maxTimestamp)
        for tier in self.tiers:
            shrunkTier = tier.eraseRegion(
                start, end, constants.EraseCollision.TRUNCATE, doShrink
            )
            newTG.addTier(shrunkTier)

        newTG.maxTimestamp = maxTimestamp

        return newTG

    def editTimestamps(
        self,
        offset: float,
        reportingMode: Literal["silence", "warning", "error"] = "warning",
    ) -> "Textgrid":
        """Modifies all timestamps by a constant amount

        Args:
            offset: the amount to offset in seconds
            reportingMode: one of "silence", "warning", or "error". This flag
                determines the behavior if there is a size difference between the
                maxTimestamp in the tier and the current textgrid.

        Returns:
            Textgrid: the modified version of the current textgrid
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )

        tg = Textgrid(self.minTimestamp, self.maxTimestamp)
        for tier in self.tiers:
            if len(tier.entries) > 0:
                tier = tier.editTimestamps(offset, reportingMode)

            tg.addTier(tier, reportingMode=reportingMode)

        return tg

    def getTier(self, tierName: str) -> textgrid_tier.TextgridTier:
        """Get the tier with the specified name"""
        return self._tierDict[tierName]

    def insertSpace(
        self,
        start: float,
        duration: float,
        collisionMode: Literal["stretch", "split", "no_change", "error"] = "error",
    ) -> "Textgrid":
        """Inserts a blank region into a textgrid

        Every item that occurs after *start* will be pushed back by
        *duration* seconds

        Args:
            start:
            duration:
            collisionMode: Determines behavior in the event that an
                interval stradles the starting point.
                One of ['stretch', 'split', 'no change', None]
                - 'stretch' stretches the interval by /duration/ amount
                - 'split' splits the interval into two--everything to the
                    right of 'start' will be advanced by 'duration' seconds
                - 'no change' leaves the interval as is with no change
                - None or any other value throws an AssertionError

        Returns:
            Textgrid: the modified version of the current textgrid
        """
        utils.validateOption(
            "collisionMode", collisionMode, constants.WhitespaceCollision
        )

        newTG = Textgrid(self.minTimestamp, self.maxTimestamp)
        newTG.minTimestamp = self.minTimestamp
        newTG.maxTimestamp = self.maxTimestamp + duration

        for tier in self.tiers:
            newTier = tier.insertSpace(start, duration, collisionMode)
            newTG.addTier(newTier)

        return newTG

    def mergeTiers(
        self, tierNames: Optional[Sequence[str]] = None, preserveOtherTiers: bool = True
    ) -> "Textgrid":
        """Combine tiers

        Args:
            tierList: A list of tier names to combine. If none, combine
                all tiers.
            preserveOtherTiers: If false, uncombined tiers are not
                included in the output.

        Returns:
            Textgrid: the modified version of the current textgrid
        """
        if tierNames is None:
            tierNames = self.tierNames

        # Determine the tiers to merge
        intervalTierNames = []
        pointTierNames = []
        for tierName in tierNames:
            tier = self.getTier(tierName)
            if isinstance(tier, interval_tier.IntervalTier):
                intervalTierNames.append(tierName)
            elif isinstance(tier, point_tier.PointTier):
                pointTierNames.append(tierName)

        # Merge the interval tiers
        intervalTier = None
        if len(intervalTierNames) > 0:
            intervalTier = self.getTier(intervalTierNames.pop(0))
            for tierName in intervalTierNames:
                intervalTier = intervalTier.union(self.getTier(tierName))

        # Merge the point tiers
        pointTier = None
        if len(pointTierNames) > 0:
            pointTier = self.getTier(pointTierNames.pop(0))
            for tierName in pointTierNames:
                pointTier = pointTier.union(self.getTier(tierName))

        # Create the final textgrid to output
        tg = Textgrid(self.minTimestamp, self.maxTimestamp)

        if preserveOtherTiers:
            for tier in self.tiers:
                if tier.name not in tierNames:
                    tg.addTier(tier)

        if intervalTier is not None:
            tg.addTier(intervalTier)

        if pointTier is not None:
            tg.addTier(pointTier)

        return tg

    def new(self) -> "Textgrid":
        """Returns a copy of this Textgrid"""
        return copy.deepcopy(self)

    def save(
        self,
        fn: str,
        format: Literal["short_textgrid", "long_textgrid", "json", "textgrid_json"],
        includeBlankSpaces: bool,
        minTimestamp: Optional[float] = None,
        maxTimestamp: Optional[float] = None,
        minimumIntervalLength: float = MIN_INTERVAL_LENGTH,
        reportingMode: Literal["silence", "warning", "error"] = "warning",
    ) -> None:
        """Save the current textgrid to a file

        Args:
            fn: the fullpath filename of the output
            format: one of ['short_textgrid', 'long_textgrid', 'json', 'textgrid_json']
                'short_textgrid' and 'long_textgrid' are both used by praat
                'json' and 'textgrid_json' are two json variants. 'json' cannot represent
                tiers with different min and max timestamps than the textgrid.
            includeBlankSpaces: if True, blank sections in interval
                tiers will be filled in with an empty interval
                (with a label of ""). If you are unsure, True is recommended
                as Praat needs blanks to render textgrids properly.
            minTimestamp: the minTimestamp of the saved Textgrid;
                if None, use whatever is defined in the Textgrid object.
                If minTimestamp is larger than timestamps in your textgrid,
                an exception will be thrown.
            maxTimestamp: the maxTimestamp of the saved Textgrid;
                if None, use whatever is defined in the Textgrid object.
                If maxTimestamp is smaller than timestamps in your textgrid,
                an exception will be thrown.
            minimumIntervalLength: any labeled intervals smaller
                than this will be removed, useful for removing ultrashort
                or fragmented intervals; if None, don't remove any.
                Removed intervals are merged (without their label) into
                adjacent entries.
            reportingMode: one of "silence", "warning", or "error". This flag
                determines the behavior if there is a size difference between the
                maxTimestamp in the tier and the current textgrid.

        Returns:
            a string representation of the textgrid
        """

        utils.validateOption("format", format, TextgridFormats)
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )

        self.validate(reportingMode)

        tgAsDict = _tgToDictionary(self)

        textgridStr = textgrid_io.getTextgridAsStr(
            tgAsDict,
            format,
            includeBlankSpaces,
            minTimestamp,
            maxTimestamp,
            minimumIntervalLength,
        )

        with io.open(fn, "w", encoding="utf-8") as fd:
            fd.write(textgridStr)

    def renameTier(self, oldName: str, newName: str) -> None:
        oldTier = self.getTier(oldName)
        tierIndex = self.tierNames.index(oldName)
        self.removeTier(oldName)
        self.addTier(oldTier.new(newName, oldTier.entries), tierIndex)

    def removeTier(self, name: str) -> textgrid_tier.TextgridTier:
        return self._tierDict.pop(name)

    def replaceTier(
        self,
        name: str,
        newTier: textgrid_tier.TextgridTier,
        reportingMode: Literal["silence", "warning", "error"] = "warning",
    ) -> None:
        tierIndex = self.tierNames.index(name)
        self.removeTier(name)
        self.addTier(newTier, tierIndex, reportingMode)

    def validate(
        self, reportingMode: Literal["silence", "warning", "error"] = "warning"
    ) -> bool:
        """Validates this textgrid

        Returns whether the textgrid is valid or not. If reportingMode is "warning"
        or "error" this will also print on error or stop execution, respectively.

        Args:
            reportingMode: one of "silence", "warning", or "error". This flag
                determines the behavior if there is a size difference between the
                maxTimestamp in a tier and the current textgrid.

        Returns:
            True if this Textgrid is valid; False if not

        Raises:
            TierNameExistsError: Two tiers have the same name
            TextgridStateError: A timestamp fall outside of the allowable range
        """
        utils.validateOption(
            "reportingMode", reportingMode, constants.ErrorReportingMode
        )
        errorReporter = utils.getErrorReporter(reportingMode)

        isValid = True
        if len(self.tierNames) != len(set(self.tierNames)):
            isValid = False
            errorReporter(
                errors.TierNameExistsError,
                f"Tier names not unique: {self.tierNames}",
            )

        for tier in self.tiers:
            if self.minTimestamp != tier.minTimestamp:
                isValid = False
                errorReporter(
                    errors.TextgridStateError,
                    f"Textgrid has a min timestamp of ({self.minTimestamp}) "
                    f"but tier has ({tier.minTimestamp})",
                )

            if self.maxTimestamp != tier.maxTimestamp:
                isValid = False
                errorReporter(
                    errors.TextgridStateError,
                    f"Textgrid has a max timestamp of ({self.maxTimestamp}) "
                    f"but tier has ({tier.maxTimestamp})",
                )

            isValid = isValid and tier.validate(reportingMode)

        return isValid


def _tgToDictionary(tg: Textgrid) -> dict:
    tiers = []
    for tier in tg.tiers:
        tierAsDict = {
            "class": tier.tierType,
            "name": tier.name,
            "xmin": tier.minTimestamp,
            "xmax": tier.maxTimestamp,
            "entries": tier.entries,
        }
        tiers.append(tierAsDict)

    return {"xmin": tg.minTimestamp, "xmax": tg.maxTimestamp, "tiers": tiers}



from typing import List

from praatio.utilities import constants


class PraatioException(Exception):
    pass


class SafeZipException(PraatioException):
    pass


class FileNotFound(PraatioException):
    def __init__(self, fullPath: str):
        super(FileNotFound, self).__init__()
        self.fullPath = fullPath

    def __str__(self):
        return "File not found:\n%s" % self.fullPath


class ParsingError(PraatioException):
    pass


class ArgumentError(PraatioException):
    pass


class UnexpectedError(PraatioException):
    pass


class WrongOption(PraatioException):
    def __init__(self, argumentName: str, givenValue: str, availableOptions: List[str]):
        self.argumentName = argumentName
        self.givenValue = givenValue
        self.availableOptions = availableOptions

    def __str__(self):
        return (
            f"For argument '{self.argumentName}' was given the value '{self.givenValue}'. "
            f"However, expected one of [{', '.join(self.availableOptions)}]"
        )


class TextgridException(PraatioException):
    pass


class DuplicateTierName(TextgridException):
    pass


class OutOfBounds(TextgridException):
    pass


class CollisionError(TextgridException):
    pass


class TimelessTextgridTierException(TextgridException):
    def __str__(self):
        return "All textgrid tiers much have a min and max duration"


# When the state of a textgrid has to change in a way the user did
# not expect (e.g. a new interval was added that is longer
# than the maxTimestamp, causing maxTimestamp to be lengthened)
class TextgridStateAutoModified(TextgridException):
    pass


class TextgridStateError(TextgridException):
    pass


class TierNameExistsError(TextgridException):
    pass


class IncompatibleTierError(TextgridException):
    def __init__(self, tier):
        super(IncompatibleTierError, self).__init__()
        self.tier = tier
        if self.tier.tierType == constants.INTERVAL_TIER:
            self.otherTierType = constants.POINT_TIER
        else:
            self.otherTierType = constants.INTERVAL_TIER

    def __str__(self):
        return (
            f"Incompatible tier type.  Tier with name {self.tier.name} has type"
            f"{self.tier.tierType} but expected {self.otherTierType}"
        )


class PraatExecutionFailed(PraatioException):
    def __init__(self, cmdList: List[str]):
        super(PraatExecutionFailed, self).__init__()
        self.cmdList = cmdList

    def __str__(self):
        errorStr = (
            "\nPraat Execution Failed.  Please check the following:\n"
            "- Praat exists in the location specified\n"
            "- Praat script can execute ok outside of praat\n"
            "- script arguments are correct\n\n"
            "If you can't locate the problem, I recommend using "
            "absolute paths rather than relative "
            "paths and using paths without spaces in any folder "
            "or file names\n\n"
            "Here is the command that python attempted to run:\n"
        )
        cmdTxt = " ".join(self.cmdList)
        return errorStr + cmdTxt


class FindZeroCrossingError(PraatioException):
    def __init__(self, startTime: float, endTime: float):
        super(FindZeroCrossingError, self).__init__()

        self.startTime = startTime
        self.endTime = endTime

    def __str__(self):
        retString = "No zero crossing found between %f and %f"
        return retString % (self.startTime, self.endTime)


class NormalizationException(PraatioException):
    def __str__(self):
        return (
            "Local normalization will nullify the effect of global normalization. "
            "Local normalization should be used to examine local phenomena "
            "(e.g. max pitch in a segment of running speech)."
            "Global normalization should be used to examine global phenomena "
            "(e.g. the pitch range of a speaker)."
        )



import re
import json
from typing import Optional, Tuple, List, Any, Dict, Match

from typing_extensions import Literal

from praatio.utilities import errors
from praatio.utilities import my_math
from praatio.utilities import utils
from praatio.utilities.constants import (
    TextgridFormats,
    MIN_INTERVAL_LENGTH,
    Interval,
    Point,
    INTERVAL_TIER,
    POINT_TIER,
)


def reSearch(pattern, string, flags=None) -> Match[str]:
    """Search for the string to match. Throws an error if no match is found."""
    if flags:
        matches = re.search(pattern, string, flags)
    else:
        matches = re.search(pattern, string)

    if not matches:
        raise errors.ParsingError("Expected field in Textgrid missing.")

    return matches


def _removeBlanks(tier: Dict) -> None:
    def hasContent(entry):
        return entry[-1] != ""

    tier["entries"] = filter(hasContent, tier["entries"])


def _removeUltrashortIntervals(
    tier: Dict, minLength: float, minTimestamp: float
) -> None:
    """Remove intervals that are very tiny

    Doing many small manipulations on intervals can lead to the creation
    of ultrashort intervals (e.g. 1*10^-15 seconds long).  This function
    removes such intervals.
    """

    # First, remove tiny intervals
    newEntries: List[Interval] = []
    j = 0  # index to newEntries
    for start, end, label in tier["entries"]:

        if end - start < minLength:
            # Correct ultra-short entries
            if len(newEntries) > 0:
                lastStart, _, lastLabel = newEntries[j - 1]
                newEntries[j - 1] = Interval(lastStart, end, lastLabel)
        else:
            # Special case: the first entry in oldEntries was ultra-short
            if len(newEntries) == 0 and start != minTimestamp:
                newEntries.append(Interval(minTimestamp, end, label))
            # Normal case
            else:
                newEntries.append(Interval(start, end, label))
            j += 1

    # Next, shift near equivalent tiny boundaries
    # This will link intervals that were connected by an interval
    # that was shorter than minLength
    j = 0
    while j < len(newEntries) - 1:
        diff = abs(newEntries[j][1] - newEntries[j + 1][0])
        if diff > 0 and diff < minLength:
            newEntries[j] = Interval(
                newEntries[j][0],
                newEntries[j + 1][0],
                newEntries[j][2],
            )
        j += 1

    tier["entries"] = newEntries


def _fillInBlanks(
    tier: Dict,
    blankLabel: str = "",
    minTime: Optional[float] = None,
    maxTime: Optional[float] = None,
) -> None:
    """Fills in the space between intervals with empty space

    This is necessary to do when saving to create a well-formed textgrid
    """
    if minTime is None:
        minTime = tier["xmin"]

    if maxTime is None:
        maxTime = tier["xmax"]

    # Special case: empty textgrid
    if len(tier["entries"]) == 0:
        tier["entries"].append((minTime, maxTime, blankLabel))

    # Create a new entry list
    entries = tier["entries"]
    entry = entries[0]
    prevEnd = float(entry[1])
    newEntries = [entry]
    for entry in entries[1:]:
        newStart = float(entry[0])
        newEnd = float(entry[1])

        if prevEnd < newStart:
            newEntries.append((prevEnd, newStart, blankLabel))
        newEntries.append(entry)

        prevEnd = newEnd

    # Special case: If there is a gap at the start of the file
    if float(newEntries[0][0]) < float(minTime):
        raise errors.ParsingError(
            "The entries are shorter than the min time specified in the textgrid."
        )
    if float(newEntries[0][0]) > float(minTime):
        newEntries.insert(0, (minTime, newEntries[0][0], blankLabel))

    # Special case -- if there is a gap at the end of the file
    if maxTime is not None:
        if float(newEntries[-1][1]) > float(maxTime):
            raise errors.ParsingError(
                "The entries are longer than the max time specified in the textgrid."
            )
        if float(newEntries[-1][1]) < float(maxTime):
            newEntries.append((newEntries[-1][1], maxTime, blankLabel))

    newEntries.sort()
    tier["entries"] = newEntries


def parseTextgridStr(data: str, includeEmptyIntervals: bool = False) -> Dict:
    """Converts a string representation of a Textgrid into a dictionary

    https://www.fon.hum.uva.nl/praat/manual/TextGrid_file_formats.html

    Args:
        fnFullPath: the path to the textgrid to open
        includeEmptyIntervals: if False, points and intervals with
             an empty label '' are not included in the returned dictionary

    Returns:
        Dictionary
    """

    try:
        tgAsDict = json.loads(data)
        if "start" in tgAsDict.keys():  # Using simplified json format
            tgAsDict = _upconvertDictionaryFromJson(tgAsDict)
    except ValueError:
        caseA = "ooTextFile short" in data
        caseB = "item [" not in data
        if caseA or caseB:
            tgAsDict = _parseShortTextgrid(data)
        else:
            tgAsDict = _parseNormalTextgrid(data)

    if includeEmptyIntervals is False:
        for tier in tgAsDict["tiers"]:
            _removeBlanks(tier)

    return tgAsDict


def getTextgridAsStr(
    tg: Dict,
    format: Literal["short_textgrid", "long_textgrid", "json", "textgrid_json"],
    includeBlankSpaces: bool,
    minTimestamp: Optional[float] = None,
    maxTimestamp: Optional[float] = None,
    minimumIntervalLength: float = MIN_INTERVAL_LENGTH,
) -> str:
    """Converts a textgrid to a string, suitable for saving

    Args:
        tg: the textgrid to convert to a string
        format: one of ['short_textgrid', 'long_textgrid', 'json', 'textgrid_json']
        includeBlankSpaces: if True, blank sections in interval
            tiers will be filled in with an empty interval
            (with a label of "")
        minTimestamp: the minTimestamp of the saved Textgrid;
            if None, use whatever is defined in the Textgrid object.
            If minTimestamp is larger than timestamps in your textgrid,
            an exception will be thrown.
        maxTimestamp: the maxTimestamp of the saved Textgrid;
            if None, use whatever is defined in the Textgrid object.
            If maxTimestamp is smaller than timestamps in your textgrid,
            an exception will be thrown.
        minimumIntervalLength: any labeled intervals smaller
            than this will be removed, useful for removing ultrashort
            or fragmented intervals; if None, don't remove any.
            Removed intervals are merged (without their label) into
            adjacent entries.

    Returns:
        a string representation of the textgrid
    """

    utils.validateOption("format", format, TextgridFormats)

    tg = _prepTgForSaving(
        tg, includeBlankSpaces, minTimestamp, maxTimestamp, minimumIntervalLength
    )

    if format == TextgridFormats.LONG_TEXTGRID:
        outputTxt = _tgToLongTextForm(tg)
    elif format == TextgridFormats.SHORT_TEXTGRID:
        outputTxt = _tgToShortTextForm(tg)
    elif format == TextgridFormats.JSON:
        outputTxt = _tgToJson(_downconvertDictionaryForJson(tg))
    elif format == TextgridFormats.TEXTGRID_JSON:
        outputTxt = _tgToJson(tg)

    return outputTxt


def _upconvertDictionaryFromJson(tgAsDict: dict) -> dict:
    """
    Convert from the sparse json format to the one shaped more literally like a textgrid
    """
    transformedDict = {}
    transformedDict["xmin"] = tgAsDict["start"]
    transformedDict["xmax"] = tgAsDict["end"]
    transformedDict["tiers"] = []

    for tierName in tgAsDict["tiers"].keys():
        tier = tgAsDict["tiers"][tierName]
        transformedDict["tiers"].append(
            {
                "class": tier["type"],
                "name": tierName,
                "xmin": tgAsDict["start"],
                "xmax": tgAsDict["end"],
                "entries": tier["entries"],
            }
        )

    return transformedDict


def _downconvertDictionaryForJson(tgAsDict: Dict) -> dict:
    """
    Convert from the textgrid-shaped json format to a more minimal json format
    """
    tiers = {}
    for tier in tgAsDict["tiers"]:
        tiers[tier["name"]] = {
            "type": tier["class"],
            "entries": tier["entries"],
        }

    return {
        "start": tgAsDict["xmin"],
        "end": tgAsDict["xmax"],
        "tiers": tiers,
    }


def _sortEntries(tg: Dict) -> None:
    for tier in tg["tiers"]:
        tier["entries"] = sorted(tier["entries"])


def _prepTgForSaving(
    tg: Dict,
    includeBlankSpaces: Optional[bool],
    minTimestamp: Optional[float],
    maxTimestamp: Optional[float],
    minimumIntervalLength: float,
) -> Dict:
    _sortEntries(tg)

    if minTimestamp is None:
        minTimestamp = tg["xmin"]
    else:
        tg["xmin"] = minTimestamp

    if maxTimestamp is None:
        maxTimestamp = tg["xmax"]
    else:
        tg["xmax"] = maxTimestamp

    # Fill in the blank spaces for interval tiers
    if includeBlankSpaces:
        newTierList = []
        for tier in tg["tiers"]:
            if tier["class"] == POINT_TIER:
                newTierList.append(tier)
                continue

            _fillInBlanks(tier, "", minTimestamp, maxTimestamp)
            if minimumIntervalLength is not None:
                _removeUltrashortIntervals(tier, minimumIntervalLength, minTimestamp)

    _sortEntries(tg)

    return tg


def _tgToShortTextForm(
    tg: Dict,
) -> str:

    # Header
    outputTxt = ""
    outputTxt += 'File type = "ooTextFile"\n'
    outputTxt += 'Object class = "TextGrid"\n\n'
    outputTxt += "%s\n%s\n" % (
        my_math.numToStr(tg["xmin"]),
        my_math.numToStr(tg["xmax"]),
    )
    outputTxt += "<exists>\n%d\n" % len(tg["tiers"])
    for tier in tg["tiers"]:
        text = ""
        text += '"%s"\n' % tier["class"]
        text += '"%s"\n' % utils.escapeQuotes(tier["name"])
        text += "%s\n%s\n%s\n" % (
            my_math.numToStr(tier["xmin"]),
            my_math.numToStr(tier["xmax"]),
            len(tier["entries"]),
        )

        for entry in tier["entries"]:
            entry = [my_math.numToStr(val) for val in entry[:-1]] + [
                '"%s"' % utils.escapeQuotes(entry[-1])
            ]

            text += "\n".join([str(val) for val in entry]) + "\n"

        outputTxt += text

    return outputTxt


def _tgToLongTextForm(tg: Dict) -> str:
    outputTxt = ""
    outputTxt += 'File type = "ooTextFile"\n'
    outputTxt += 'Object class = "TextGrid"\n\n'

    tab = " " * 4

    # Header
    outputTxt += "xmin = %s \n" % my_math.numToStr(tg["xmin"])
    outputTxt += "xmax = %s \n" % my_math.numToStr(tg["xmax"])
    outputTxt += "tiers? <exists> \n"
    outputTxt += "size = %d \n" % len(tg["tiers"])
    outputTxt += "item []: \n"

    for tierNum, tier in enumerate(tg["tiers"]):
        # Interval header
        outputTxt += tab + "item [%d]:\n" % (tierNum + 1)
        outputTxt += tab * 2 + 'class = "%s" \n' % tier["class"]
        outputTxt += tab * 2 + 'name = "%s" \n' % utils.escapeQuotes(tier["name"])
        outputTxt += tab * 2 + "xmin = %s \n" % my_math.numToStr(tier["xmin"])
        outputTxt += tab * 2 + "xmax = %s \n" % my_math.numToStr(tier["xmax"])

        entries = tier["entries"]
        if tier["class"] == INTERVAL_TIER:
            outputTxt += tab * 2 + "intervals: size = %d \n" % len(entries)
            for intervalNum, entry in enumerate(entries):
                start, end, label = entry
                outputTxt += tab * 2 + "intervals [%d]:\n" % (intervalNum + 1)
                outputTxt += tab * 3 + "xmin = %s \n" % my_math.numToStr(start)
                outputTxt += tab * 3 + "xmax = %s \n" % my_math.numToStr(end)
                outputTxt += tab * 3 + 'text = "%s" \n' % utils.escapeQuotes(label)
        else:
            outputTxt += tab * 2 + "points: size = %d \n" % len(entries)
            for pointNum, entry in enumerate(entries):
                timestamp, label = entry
                outputTxt += tab * 2 + "points [%d]:\n" % (pointNum + 1)
                outputTxt += tab * 3 + "number = %s \n" % my_math.numToStr(timestamp)
                outputTxt += tab * 3 + 'mark = "%s" \n' % utils.escapeQuotes(label)

    return outputTxt


def _tgToJson(tgAsDict: Dict) -> str:
    """Returns a json representation of a textgrid"""
    return json.dumps(tgAsDict, ensure_ascii=False)


def _parseNormalTextgrid(data: str) -> Dict:
    """
    Reads a normal textgrid
    """
    data = data.replace("\r\n", "\n")

    # Toss textgrid header
    header, data = re.split(r"item ?\[", data, maxsplit=1, flags=re.MULTILINE)

    headerList = header.split("\n")
    tgMin = float(headerList[3].split("=")[1].strip())
    tgMax = float(headerList[4].split("=")[1].strip())

    # Process each tier individually (will be output to separate folders)
    tiers = []
    tierList = re.split(r"item ?\[", data, flags=re.MULTILINE)[1:]
    for tierTxt in tierList:
        if 'class = "IntervalTier"' in tierTxt:
            tierType = INTERVAL_TIER
            searchWord = r"intervals ?\["
        else:
            tierType = POINT_TIER
            searchWord = r"points ?\["

        # Get tier meta-information
        try:
            d = re.split(searchWord, tierTxt, flags=re.MULTILINE)
            header, tierData = d[0], d[1:]
        except ValueError:
            # A tier with no entries
            if re.search(r"size ?= ?0", tierTxt):
                header = tierTxt
                tierData = []
            else:
                raise
        tierName = reSearch(
            r"name ?= ?\"(.*)\"\s*$", header, flags=re.MULTILINE
        ).groups()[0]
        tierName = re.sub(r'""', '"', tierName)

        # "-0" has been reported as a potential start time
        tierStartTimeStr = reSearch(
            r"xmin ?= ?-?([\d.]+)\s*$", header, flags=re.MULTILINE
        ).groups()[0]
        tierStartTime = utils.strToIntOrFloat(tierStartTimeStr)

        tierEndTimeStr = reSearch(
            r"xmax ?= ?([\d.]+)\s*$", header, flags=re.MULTILINE
        ).groups()[0]
        tierEndTime = utils.strToIntOrFloat(tierEndTimeStr)

        # Get the tier entry list
        entries: List[Any] = []
        if tierType == INTERVAL_TIER:
            for element in tierData:
                timeStart = reSearch(
                    r"xmin ?= ?-?([\d.]+)\s*$", element, flags=re.MULTILINE
                ).groups()[0]
                timeEnd = reSearch(
                    r"xmax ?= ?([\d.]+)\s*$", element, flags=re.MULTILINE
                ).groups()[0]
                label = reSearch(
                    r"text ?= ?\"(.*)\"\s*$",
                    element,
                    flags=re.MULTILINE | re.DOTALL,
                ).groups()[0]

                label = label.strip()
                label = re.sub(r'""', '"', label)
                entries.append(Interval(timeStart, timeEnd, label))
        else:
            for element in tierData:
                time = reSearch(
                    r"number ?= ?-?([\d.]+)\s*$", element, flags=re.MULTILINE
                ).groups()[0]
                label = reSearch(
                    r"mark ?= ?\"(.*)\"\s*$",
                    element,
                    flags=re.MULTILINE | re.DOTALL,
                ).groups()[0]
                label = label.strip()
                entries.append(Point(time, label))

        tierAsDict = {
            "class": tierType,
            "name": tierName,
            "xmin": float(tierStartTime),
            "xmax": float(tierEndTime),
            "entries": entries,
        }
        tiers.append(tierAsDict)

    tgDict = {"xmin": tgMin, "xmax": tgMax, "tiers": tiers}

    return tgDict


def _parseShortTextgrid(data: str) -> Dict:
    """Reads a short textgrid file"""
    data = data.replace("\r\n", "\n")

    intervalIndicies = [(i, True) for i in utils.findAll(data, '"IntervalTier"')]
    pointIndicies = [(i, False) for i in utils.findAll(data, '"TextTier"')]

    indexList = [*intervalIndicies, *pointIndicies]
    indexList.append((len(data), True))  # The 'end' of the file
    indexList.sort()

    tupleList = [
        (indexList[i][0], indexList[i + 1][0], indexList[i][1])
        for i in range(len(indexList) - 1)
    ]

    # Set the textgrid's min and max times
    header = data[: tupleList[0][0]]
    headerList = header.split("\n")
    tgMin = float(headerList[3].strip())
    tgMax = float(headerList[4].strip())

    # Load the data for each tier
    tiers = []
    for blockStartI, blockEndI, isInterval in tupleList:
        tierData = data[blockStartI:blockEndI]

        # First row contains the tier type, which we already know
        metaStartI = _fetchRow(tierData, 0)[1]

        # Tier meta-information
        tierName, tierNameEndI = _fetchTextRow(tierData, metaStartI)
        tierStartTimeStr, tierStartTimeI = _fetchRow(tierData, tierNameEndI)
        tierEndTimeStr, tierEndTimeI = _fetchRow(tierData, tierStartTimeI)
        startTimeI = _fetchRow(tierData, tierEndTimeI)[1]

        tierStartTime = utils.strToIntOrFloat(tierStartTimeStr)
        tierEndTime = utils.strToIntOrFloat(tierEndTimeStr)

        # Tier entry data
        entries: List[Any] = []
        if isInterval:
            className = INTERVAL_TIER
            while True:
                try:
                    startTime, endTimeI = _fetchRow(tierData, startTimeI)
                    endTime, labelI = _fetchRow(tierData, endTimeI)
                    label, startTimeI = _fetchTextRow(tierData, labelI)
                except (ValueError, IndexError):
                    break

                label = label.strip()
                entries.append(Interval(startTime, endTime, label))
        else:
            className = POINT_TIER
            while True:
                try:
                    time, labelI = _fetchRow(tierData, startTimeI)
                    label, startTimeI = _fetchTextRow(tierData, labelI)
                except (ValueError, IndexError):
                    break
                label = label.strip()
                entries.append(Point(time, label))

        tierAsDict = {
            "class": className,
            "name": tierName,
            "xmin": float(tierStartTime),
            "xmax": float(tierEndTime),
            "entries": entries,
        }
        tiers.append(tierAsDict)

    tgDict = {"xmin": tgMin, "xmax": tgMax, "tiers": tiers}

    return tgDict


def _fetchRow(
    dataStr: str, index: int, searchStr: Optional[str] = None
) -> Tuple[str, int]:
    if searchStr is None:
        startIndex = index
    else:
        startIndex = dataStr.index(searchStr, index) + len(searchStr)

    endIndex = dataStr.index("\n", startIndex)

    word = dataStr[startIndex:endIndex]
    word = word.strip()
    if word[0] == '"' and word[-1] == '"':
        word = word[1:-1]
    word = word.strip()

    return word, endIndex + 1


def _fetchTextRow(
    dataStr: str, index: int, searchStr: Optional[str] = None
) -> Tuple[str, int]:
    if searchStr is None:
        startIndex = index
    else:
        startIndex = dataStr.index(searchStr, index) + len(searchStr)

    # A textgrid text is ended by double quotes. Double quotes that
    # appear in the text are escaped by a preceeding double quotes.
    # We know we're at the end of a text if the number of double
    # quotes is odd.
    endIndex = startIndex + 1
    while True:
        quoteStartIndex = dataStr.index('"', endIndex)
        quoteEndIndex = quoteStartIndex
        while dataStr[quoteEndIndex] == '"':
            quoteEndIndex += 1

        endIndex = quoteEndIndex

        if (quoteEndIndex - quoteStartIndex) % 2 != 0:
            break

    word = dataStr[startIndex:endIndex]
    word = word[1:-1]  # Remove the quote marks around the text
    word = word.strip()

    word = word.replace('""', '"')  # Unescape quote marks

    # Advance to the end of the line
    endIndex = dataStr.index("\n", endIndex)

    return word, endIndex + 1
